from typing import Any

from flask import Response, jsonify

from mmpm.log.factory import MMPMLogFactory

logger = MMPMLogFactory.get_logger(__name__)


class Endpoint:
    """
    A class representing a web endpoint in a Flask application. It provides methods
    to generate standard success and failure responses.
    """

    def __init__(self):
        """
        Initializes the Endpoint instance with an empty name.
        """
        self.name: str = ""

    def success(self, msg: Any) -> Response:
        """
        Generates a success response for the endpoint.

        Parameters:
            msg (Any): The message or data to be included in the response.

        Returns:
            Response: A Flask Response object with a 200 status code and the provided message.
        """
        return jsonify({"code": 200, "message": msg})

    def failure(self, msg: Any, code: int = 500) -> Response:
        """
        Generates a failure response for the endpoint.

        Parameters:
            msg (Any): The error message or data to be included in the response.
            code (int, optional): The HTTP status code for the response. Defaults to 500.

        Returns:
            Response: A Flask Response object with the specified status code and the provided message.
        """
        return jsonify({"code": code, "message": msg})



from gevent import monkey

monkey.patch_all()

import json

from flask import Flask, Response
from flask_cors import CORS

import mmpm.api.endpoints
from mmpm.api.endpoints.index import Index
from mmpm.log.factory import MMPMLogFactory
from mmpm.subcommands.loader import Loader

logger = MMPMLogFactory.get_logger(__name__)

app = Flask(__name__)
app.config["CORS_HEADERS"] = "Content-Type"
CORS(app)

resources: dict = {
    r"/*": {"origins": "*"},
    r"/api/*": {"origins": "*"},
    r"/socket.io/*": {"origins": "*"},
}


@app.after_request  # type: ignore
def after_request(response: Response) -> Response:
    """
    Appends extra headers after each api request is sent to the server

    Parameters:
        response (flask.Response): the response object being returned to the frontend

    Returns
        response (flask.Response): the modified response object with new headers attached
    """

    response.headers.add("Access-Control-Allow-Origin", "*")
    response.headers.add("Access-Control-Allow-Headers", "*")
    response.headers.add("Access-Control-Allow-Methods", "*")
    return response


@app.errorhandler(Exception)
def exception_handler(error) -> Response:
    response: Response = error.get_response()
    response.data = json.dumps({"code": error.code, "message": error.description})
    response.content_type = "application/json"
    logger.error(error.description)
    return response


# dynamically load all the endpoints within the "mmpm.api.endpoints" module
loader = Loader(
    module_path=mmpm.api.endpoints.__path__,
    module_name="mmpm.api.endpoints",
    prefix="ep_",
)

app.url_map.strict_slashes = False

entrypoints = list(loader.objects.values())
entrypoints.append(Index(app.url_map))

for endpoint in entrypoints:
    try:
        app.register_blueprint(endpoint.blueprint)  # type: ignore
        logger.debug(f"Loaded blueprint for {endpoint}")
    except Exception as exception:
        logger.error(f"Failed to load blueprint for {endpoint}: {exception}")



"""Command line options for 'guided-setup' subcommand"""

import json
from os import getenv
from pathlib import Path

from mmpm.constants import color, paths
from mmpm.env import MMPMEnv
from mmpm.log.factory import MMPMLogFactory
from mmpm.subcommands.sub_cmd import SubCmd
from mmpm.utils import confirm, get_host_ip, prompt

logger = MMPMLogFactory.get_logger(__name__)


class GuidedSetup(SubCmd):
    """
    The 'GuidedSetup' subcommand interactively walks users through the setup
    process of MMPM's features, and generates the next steps they should
    manually perform (if necessary)

    Custom Attributes:
        env (MMPMEnv): A singleton of MMPMEnv which contains environment variables
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "guided-setup"
        self.help = f"Interactively setup {self.app_name} and its features"
        self.usage = f"{self.app_name} {self.name}"
        self.env = MMPMEnv()

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

    def exec(self, args, extra):
        """
        Provides the user a guided configuration of the environment variables, and
        feature installation. This can be re-run as many times as necessary.

        Parameters:
            None

        Returns:
            None
        """
        print(color.b_green("Welcome to MMPM's guided setup!\n"))
        print("I'll help you setup your environment variables and additional features. Let's get started!\n")

        magicmirror_root: str = f"{Path.home()}/MagicMirror"
        magicmirror_uri: str = f"http://{get_host_ip()}:8080"
        magicmirror_pm2_proc: str = ""
        magicmirror_docker_compose_file: str = ""
        mmpm_is_docker_image: bool = False
        install_ui: bool = False
        install_autocomplete: bool = False
        install_as_module: bool = False

        magicmirror_root = prompt("Enter the absolute path to your MagicMirror installation: ", default=f"{Path.home()}/MagicMirror")
        mmpm_is_docker_image = confirm("Is MMPM running as a Docker image?")

        if not mmpm_is_docker_image and confirm("Did you install MagicMirror using docker-compose?"):
            magicmirror_docker_compose_file = prompt(
                f"What is the absolute path to the MagicMirror docker-compose file (ie. {Path.home()}/docker-compose.yml)?"
            )

        if not mmpm_is_docker_image and not magicmirror_docker_compose_file and confirm("Are you using PM2 to start/stop MagicMirror?"):
            magicmirror_pm2_proc = prompt("What is the name of the PM2 process for MagicMirror?")

        magicmirror_uri = prompt("Enter the address and port used to access MagicMirror: ", default=f"http://{get_host_ip()}:8080")

        install_ui = not mmpm_is_docker_image and confirm("Would you like to install the MMPM UI (user interface)?")
        install_as_module = confirm("Would you like to hide/show MagicMirror modules through MMPM?")
        install_autocomplete = confirm("Would you like to install tab-autocomplete for the MMPM CLI?")

        with open(paths.MMPM_ENV_FILE, "w", encoding="utf-8") as env:
            json.dump(
                {
                    self.env.MMPM_MAGICMIRROR_ROOT.name: str(magicmirror_root),
                    self.env.MMPM_MAGICMIRROR_URI.name: magicmirror_uri,
                    self.env.MMPM_MAGICMIRROR_PM2_PROCESS_NAME.name: magicmirror_pm2_proc,
                    self.env.MMPM_MAGICMIRROR_DOCKER_COMPOSE_FILE.name: str(magicmirror_docker_compose_file),
                    self.env.MMPM_IS_DOCKER_IMAGE.name: bool(mmpm_is_docker_image),
                },
                env,
                indent=2,
            )

        message = "Based on your responses, your environment variables have been set as:"
        line_break = color.b_green("-" * len(message))

        print("\n" + line_break)
        print("\nBased on your responses, your environment variables have been set as:")
        self.env.display()

        if install_as_module or install_ui or install_autocomplete:
            print("Execute the following commands to install the desired features:")

        if install_as_module:
            print(color.b_green("mmpm install -y MMM-mmpm"))
        if install_ui:
            print(color.b_green("mmpm ui install -y"))
        if install_autocomplete:
            print(color.b_green(f"mmpm completion --shell={Path(getenv('SHELL')).stem}"))



"""
An incredibly simplistic SocketIO server used for repeating logs from the MMPM CLI to the UI.
"""

from gevent import monkey

monkey.patch_all()

import socketio

from mmpm.log.factory import MMPMLogFactory

logger = MMPMLogFactory.get_logger(__name__)


def create():
    server = socketio.Server(cors_allowed_origins="*", async_mode="gevent")

    @server.event
    def connect(sid, environ):  # pylint: disable=unused-argument
        logger.debug(f"Client connected to SocketIO-Log-Server: {sid}")

    @server.event
    def logs(sid, data):
        server.emit("logs", data, skip_sid=sid)

    @server.event
    def disconnect(sid):
        logger.debug("Client disconnected:", sid)

    app = socketio.WSGIApp(server)
    return app



import json
import os
import socket
import subprocess
import time
import urllib.request
from pathlib import Path
from typing import List, Tuple

import git
import requests
from packaging import version
from prompt_toolkit import prompt as ptk_prompt
from prompt_toolkit.shortcuts import confirm as ptk_confirm
from yaspin import yaspin
from yaspin.spinners import Spinners

from mmpm.__version__ import version as current_version
from mmpm.constants import color
from mmpm.log.factory import MMPMLogFactory

logger = MMPMLogFactory.get_logger(__name__)


def repo_up_to_date(path: Path):
    """
    Checks if the Git repository at the given path is up-to-date with its remote origin.

    Parameters:
        path (Path): The file system path to the Git repository.

    Returns:
        bool: True if the local repository is up-to-date, False otherwise.
    """

    try:
        repo = git.Repo(path)

        # Ensure the repository is not bare (unlikely, but still should check)
        if repo.bare:
            logger.error(f"Repository in {path} is bare. Cannot determine if out-of-date.")
            return False

        logger.debug(f"Fetching information for repo found in '{path}'")
        remote = repo.remotes.origin
        remote.fetch()

        # Get local and remote HEAD commit
        local_commit = repo.head.commit
        remote_commit = repo.refs["origin/HEAD"].commit  # type: ignore

        logger.debug(f"SHAs found in '{path}' -- local={local_commit.hexsha} & remote={remote_commit.hexsha}")

        remote = repo.remotes.origin
        return local_commit.hexsha != remote_commit.hexsha
    except Exception as error:
        logger.error(f"Failed to get status of repo located at {path}: {error}")
        return False


def get_host_ip() -> str:
    """
    Retrieves the local IP address of the host machine.

    Returns:
        str: The local IP address.
    """

    logger.debug("Getting host IP")

    address = "localhost"
    skt = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    try:
        skt.connect(("8.8.8.8", 80))
        address = skt.getsockname()[0]
        logger.debug(f"Determined Host IP={address}")
    except socket.gaierror as error:
        logger.error(f"Failed to determine host IP address: {error}")
        address = "localhost"  # just to be extra safe and make sure it's set to something usable
    finally:
        skt.close()

    return address


def run_cmd(command: List[str], progress=True, background=False, message: str = "") -> Tuple[int, str, str]:
    """
    Executes a shell command and captures its output and errors.

    Parameters:
        command (List[str]): The command and its arguments to be executed.
        progress (bool): If True, displays a spinner during command execution.
        background (bool): If True, runs the command in the background.
        message (str): The message to display alongside the spinner.

    Returns:
        Tuple[int, str, str]: A tuple containing the command's return code, standard output, and standard error.
    """
    if background:
        logger.debug(f"Executing command `{' '.join(command)}` in background")

        # fully detach the terminal from the process so nothing hangs
        with open(os.devnull, "wb") as devnull:
            # pylint: disable=subprocess-popen-preexec-fn
            subprocess.Popen(command, stdout=devnull, stderr=devnull, stdin=devnull, close_fds=True, preexec_fn=os.setsid)

        return 0, "", ""

    logger.debug(f"Executing command `{' '.join(command)}`")

    with subprocess.Popen(command, stderr=subprocess.PIPE, stdout=subprocess.PIPE) as process:
        if progress:
            with yaspin(text=message, color="green") as spinner:
                spinner.spinner = Spinners.bouncingBar

                while process.poll() is None:
                    time.sleep(0.1)

        stdout, stderr = process.communicate()

        return process.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")


def get_pids(process_name: str) -> List[str]:
    """
    Retrieves process IDs for all processes with the given name.

    Parameters:
        process_name (str): The name of the process to search for.

    Returns:
        List[str]: A list of process IDs.
    """

    logger.info(f"Getting process IDs related to '{process_name}'")

    with subprocess.Popen(["pgrep", process_name], stdout=subprocess.PIPE, stderr=subprocess.PIPE) as pids:
        stdout, _ = pids.communicate()
        processes = stdout.decode("utf-8")

        logger.debug(f"Found process IDs: {processes}")

        return [proc_id for proc_id in processes.split("\n") if proc_id]


def kill_pids_of_process(process: str) -> None:
    """
    Terminates all processes with the given name.

    Parameters:
        process (str): The name of the process to be terminated.
    """
    os.system(f"for process in $(pgrep {process}); do kill -9 $process; done")
    logger.debug(f"Stopped all processes of type {process}")


def safe_get_request(url: str) -> requests.Response:
    """
    Safely performs a GET request to the specified URL, handling any exceptions.

    Parameters:
        url (str): The URL to send the GET request to.

    Returns:
        requests.Response: The response from the GET request.
    """
    try:
        logger.debug(f"Creating request for {url}")
        data = requests.get(url, timeout=10)
    except requests.exceptions.RequestException as error:
        logger.error(str(error))
        return requests.Response()
    return data


def upgrade() -> bool:
    """
    Attempts to upgrade the MMPM package using pip.

    Returns:
        bool: True if the upgrade is successful, False otherwise.
    """

    error_code, stdout, stderr = run_cmd(
        ["python3", "-m", "pip", "install", "--upgrade", "mmpm"],
        message="Upgrading MMPM",
    )

    if error_code:
        logger.error(stderr)
        return False

    logger.debug(stdout)
    return True


def update_available() -> bool:
    """
    Checks if an update is available for MMPM on PyPi.

    Returns:
        bool: True if an update is available, False otherwise.
    """

    url = "https://pypi.org/pypi/mmpm/json"

    logger.debug("Getting remote version of MMPM from PyPi")
    print(f"Retrieving: {url} [{color.n_cyan('mmpm')}]")

    try:
        contents = urllib.request.urlopen(url).read()
        remote_version = json.loads(contents)["info"]["version"]
        logger.debug(f"Found remote={remote_version} & installed={current_version}")
    except (json.JSONDecodeError, urllib.error.URLError, Exception) as error:
        logger.error(f"Failed to get remote version of MMPM: {error}")

    return version.parse(remote_version) > version.parse(current_version)


# wrapping prompt_toolkit so it's easier to switch out in the future if desired
def confirm(message: str) -> bool:  # pragma: no cover
    """
    Displays a confirmation prompt to the user with the given message.

    Parameters:
        message (str): The message to display in the confirmation prompt.

    Returns:
        bool: True if the user confirms, False otherwise.
    """

    return ptk_confirm(message)


# wrapping prompt_toolkit so it's easier to switch out in the future if desired
def prompt(message: str, default=""):  # pragma: no cover
    """
    Displays a prompt to the user with the given message and waits for input.

    Parameters:
        message (str): The message to display in the prompt.

    Returns:
        str: The user's input as a string.
    """
    return ptk_prompt(message, default=default)



"""Command line options for 'mm-ctl' subcommand"""

from mmpm.env import MMPMEnv
from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.controller import MagicMirrorController
from mmpm.magicmirror.magicmirror import MagicMirror
from mmpm.subcommands.sub_cmd import SubCmd
from mmpm.utils import confirm

logger = MMPMLogFactory.get_logger(__name__)


class MmCtl(SubCmd):
    """
    The 'MmCtl' subcommand allows users to interact with the MagicMirror itself
    by hiding/showing modules, getting information about which are visible,
    even starting/stopping/restarting MagicMirror itself.

    Custom Attributes:
        database (MagicMirrorDatabase): An instance of the MagicMirrorDatabase class for managing the database.
        controller (MagicMirrorController): An instance of the MagicMirrorController for interacting with MagicMirror
        env (MMPMEnv): A singleton of MMPMEnv which contains environment variables
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "mm-ctl"
        self.help = "Commands to interact with/control MagicMirror"
        self.usage = f"{self.app_name} {self.name} [--<option>]"
        self.controller = MagicMirrorController()
        self.magicmirror = MagicMirror()
        self.env = MMPMEnv()

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

        subparsers = self.parser.add_subparsers(
            dest="command",
            description=f"use `{self.app_name} {self.name} <add/remove> --help` to see more details",
            title=f"{self.app_name} {self.name} subcommands",
            metavar="",
        )

        install_parser = subparsers.add_parser(
            "install",
            help="Install MagicMirror",
            usage=f"{self.app_name} {self.name} install [--yes]",
        )

        install_parser.add_argument(
            "-y",
            "--yes",
            action="store_true",
            help="assume yes for user response and do not show prompt",
            dest="assume_yes",
        )

        remove_parser = subparsers.add_parser(
            "remove",
            help="Remove MagicMirror",
            usage=f"{self.app_name} {self.name} remove [--yes]",
        )

        remove_parser.add_argument(
            "-y",
            "--yes",
            action="store_true",
            help="assume yes for user response and do not show prompt",
            dest="assume_yes",
        )

        self.parser.add_argument(
            "--status",
            action="store_true",
            help="show the hidden/visible status and key(s) of module(s) on your MagicMirror",
            dest="status",
        )

        self.parser.add_argument(
            "--hide",
            nargs="+",
            help="hide module(s) on your MagicMirror via provided key(s)",
            dest="hide",
        )

        self.parser.add_argument(
            "--show",
            nargs="+",
            help="show module(s) on your MagicMirror via provided key(s)",
            dest="show",
        )

        self.parser.add_argument(
            "--start",
            action="store_true",
            help="start MagicMirror; works with pm2 and docker-compose",
            dest="start",
        )

        self.parser.add_argument(
            "--stop",
            action="store_true",
            help="stop MagicMirror; works with pm2 and docker-compose",
            dest="stop",
        )

        self.parser.add_argument(
            "--restart",
            action="store_true",
            help="restart MagicMirror; works with pm2 and docker-compose",
            dest="restart",
        )

    def exec(self, args, extra):
        if extra:
            logger.error(f"Extra arguments are not accepted. See '{self.app_name} {self.name} --help'")
        elif args.command == "install":
            if confirm("Are your sure you want to install MagicMirror?"):
                self.magicmirror.install()
        elif args.command == "remove":
            if confirm("Are your sure you want to remove MagicMirror?"):
                self.magicmirror.remove()
        elif args.status:
            self.controller.status()
        elif args.hide:
            self.controller.hide(args.hide)
        elif args.show:
            self.controller.show(args.show)
        elif args.start:
            if self.env.MMPM_IS_DOCKER_IMAGE.get():
                logger.fatal("Cannot execute this command within a docker image")
            else:
                self.controller.start()
        elif args.stop:
            if self.env.MMPM_IS_DOCKER_IMAGE.get():
                logger.fatal("Cannot execute this command within a docker image")
            else:
                self.controller.stop()
        elif args.restart:
            if self.env.MMPM_IS_DOCKER_IMAGE.get():
                logger.fatal("Cannot execute this command within a docker image")
            else:
                self.controller.restart()
        else:
            logger.error(f"No arguments provided. See '{self.app_name} {self.name} --help'")



"""
This socketio server/client repeats the information collected by the MMM-mmpm module back to the UI. MagicMirror
has network settings which prevent the UI from directly interacting with MMM-mmpm if you open up the UI on a computer
other than the one hosting MMPM. It's weird, it's annoying, but it's the workaround for it. Maybe there will be simpler
solution later on, but for now this is what works. Otherwise, users would need to be told how to modify MagicMirror's
source code, and let's face it, practically nobody is going to want to do that. The other option is to make a PR to MagicMirror
but I'm sure it won't get accepted for security purposes.
"""

from gevent import monkey

monkey.patch_all()

from time import sleep

import socketio

from mmpm.env import MMPMEnv
from mmpm.log.factory import MMPMLogFactory

logger = MMPMLogFactory.get_logger(__name__)


def create():
    server = socketio.Server(cors_allowed_origins="*", async_mode="gevent", logger=logger)
    app = socketio.WSGIApp(server)
    env = MMPMEnv()
    mm_client = socketio.Client(request_timeout=300, logger=logger)
    client_ids = set()

    def setup_mm_client():
        attempt = 0
        max_retries = 250

        logger.debug("Attempting to connect to MagicMirror SocketIO server. Using a maximum of 250 retries")

        while attempt < max_retries and not mm_client.connected:
            try:
                mm_client.connect(env.MMPM_MAGICMIRROR_URI.get(), wait_timeout=10, wait=True)
                logger.debug("Successfully connected to the MagicMirror SocketIO server")
                break  # Connection successful, break out of the loop
            except Exception:
                attempt += 1
                sleep(1)

        if attempt == max_retries:
            logger.error("Maximum retry attempts reached, failed to connect.")

    @server.on("reconnect")
    @server.on("connect")
    def connect(sid, environ):  # pylint: disable=unused-argument
        logger.debug(f"Client connected to SocketIO-Repeater: {sid}")

        if sid not in client_ids:
            logger.debug(f"Added {sid} to known Client IDs")
            client_ids.add(sid)

        if not mm_client.connected:
            setup_mm_client()

    @server.event
    def disconnect(sid):
        logger.debug(f"Client disconnected: {sid}")

        if sid in client_ids:
            logger.debug(f"Removed {sid} from known Client IDs")
            client_ids.remove(sid)

        if not client_ids:
            logger.debug("No clients connected. Disconnecting from MagicMirror SocketIO Server.")
            mm_client.disconnect()

    @server.event
    def request_modules(sid):  # pylint: disable=unused-argument
        logger.debug(f"SocketIO-Repeater connected to MagicMirror SocketIO server: {mm_client.connected}")
        mm_client.emit("FROM_MMPM_APP_get_active_modules", namespace="/MMM-mmpm", data={})
        logger.debug("Emitted data request to MagicMirror SocketIO server")

    @server.event
    def error(sid):
        logger.debug(f"Client encountered error: {sid}")

    @mm_client.on("ACTIVE_MODULES", namespace="/MMM-mmpm")
    def active_modules(data):
        logger.debug(f"Received data from MagicMirror SocketIO Server: {data}")

        for client_id in client_ids:
            logger.debug(f"Repeating data to {client_id}")
            server.emit("modules", data=data, to=client_id)

    @mm_client.on("MODULES_TOGGLED", namespace="/MMM-mmpm")
    def modules_toggled(data):
        logger.debug(f"Received toggled modules from MagicMirror SocketIO Server: {data}")

        for client_id in client_ids:
            logger.debug(f"Repeating data to {client_id}")
            server.emit("modules", data=data, to=client_id)

    return app



"""Command line options for 'mm-pkg' subcommand"""

from mmpm.constants import color
from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.database import MagicMirrorDatabase
from mmpm.subcommands.sub_cmd import SubCmd
from mmpm.utils import confirm, prompt

logger = MMPMLogFactory.get_logger(__name__)


class MmPkg(SubCmd):
    """
    The 'MmPkg' subcommand allows users to add/remove custom MagicMirror packages to/from their local database

    Custom Attributes:
        database (MagicMirrorDatabase): An instance of the MagicMirrorDatabase class for managing the database.
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "mm-pkg"
        self.help = "Manually add/remove custom MagicMirror packages in your local database (similar to add-apt-repository)"
        self.usage = f"{self.app_name} {self.name} <add/remove> [--<option>]"
        self.database = MagicMirrorDatabase()

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

        subparsers = self.parser.add_subparsers(
            dest="command",
            description=f"use `{self.app_name} {self.name} <add/remove> --help` to see more details",
            title=f"{self.app_name} {self.name} subcommands",
            metavar="",
        )

        add_parser = subparsers.add_parser(
            "add",
            help="Add a custom MagicMirror package to the local database",
            usage=f"{self.app_name} {self.name} add -t <title> -a <author> -r <repo> -d <description>",
        )

        add_parser.add_argument(
            "-t",
            "--title",
            type=str,
            help="title of MagicMirror package",
            dest="title",
        )

        add_parser.add_argument(
            "-a",
            "--author",
            type=str,
            help="author of MagicMirror package",
            dest="author",
        )

        add_parser.add_argument(
            "-r",
            "--repo",
            type=str,
            help="repo URL of MagicMirror package",
            dest="repo",
        )

        add_parser.add_argument(
            "-d",
            "--desc",
            type=str,
            help="description of MagicMirror package",
            dest="desc",
        )

        remove_parser = subparsers.add_parser(
            "remove",
            help="Remove a custom MagicMirror package from the local database",
            usage=f"{self.app_name} {self.name} remove <package(s)> [--yes]",
        )

        remove_parser.add_argument(
            "pkg_name",
            nargs="+",
            help="name(s) of the MagicMirror package(s) to remove",
        )

        remove_parser.add_argument(
            "-y",
            "--yes",
            action="store_true",
            default=False,
            help="assume yes for user response and do not show prompt",
            dest="assume_yes",
        )

    def exec(self, args, extra):
        if not self.database.is_initialized():
            self.database.load()

        if args.command == "add":
            if not args.title:
                args.title = prompt("Title: ")
            if not args.author:
                args.author = prompt("Author: ")
            if not args.repo:
                args.repo = prompt("Repository: ")
            if not args.desc:
                args.desc = prompt("Description: ")

            self.database.add_mm_pkg(args.title, args.author, args.repo, args.desc)
        elif args.command == "remove":
            for name in args.pkg_name:
                if not args.assume_yes and not confirm(f"Remove {name} from the local database?"):
                    continue
                if not self.database.remove_mm_pkg(name):
                    logger.error(f"Unable to locate Custom Package named '{color.n_green(name)}'")
                else:
                    logger.info(f"Removed {color.n_green(name)}")
        else:
            logger.error(f"Invalid subcommand. See '{self.app_name} {self.name} --help'")



# pylint: disable=unused-import

"""
This script is the main entry point for a Flask-based web application. It uses gevent for asynchronous operations
and initializes the web server with specific configurations.

The gevent's monkey patching is applied at the beginning of the script to ensure that all the standard library
modules that the application uses are cooperative with greenlets. This patching is crucial for enabling asynchronous
I/O operations, which are essential for handling multiple simultaneous client requests efficiently.

The application is configured to monitor changes in several key files and reload if any of them change. This feature
is useful for development, as it allows for changes in the application's configuration or data files to be reflected
immediately without needing a manual restart.

Attributes:
    app (Flask): The Flask application instance imported from mmpm.api.entrypoint.
    logger (Logger): A logger instance for logging information, imported from mmpm.log.factory.

Functions:
    __main__: If this script is run as the main program, it starts the Flask application with specific configurations.
"""

from gevent import monkey

monkey.patch_all()  # do not move these

from mmpm.api.entrypoint import app
from mmpm.constants import paths
from mmpm.log.factory import MMPMLogFactory

logger = MMPMLogFactory.get_logger(__name__)

if __name__ == "__main__":
    app.run(
        threaded=False,
        keepalive=True,
        log=logger,
        extra_files=[
            paths.MAGICMIRROR_3RD_PARTY_PACKAGES_DB_FILE,
            paths.MMPM_CUSTOM_PACKAGES_FILE,
            paths.MMPM_AVAILABLE_UPGRADES_FILE,
            paths.MMPM_ENV_FILE,
        ],
    )



from argparse import Namespace, _SubParsersAction
from typing import List

from mmpm.log.factory import MMPMLogFactory

logger = MMPMLogFactory.get_logger(__name__)


class SubCmd:
    """
    Base subcommand class used to define all future subcommands

    Args:
        app_name (str): The name of the application.
    Attributes:
        app_name (str): The name of the application.
        name (str): The name of the subcommand.
        help (str): A brief description of the subcommand's purpose.
        usage (str): The usage information for the subcommand.
        parser (argparse._SubParsersAction): the subparser options/commands are appended to
        [Method] register(self, subparser): see method docs
        [Method] exec(self, args, extra): see method docs
    """

    def __init__(self, app_name: str = ""):
        self.app_name = app_name
        self.name: str = ""
        self.help: str = ""
        self.usage: str = ""
        self.parser = None

        if not self.name or not self.help or not self.usage or not self.parser:
            message = "name, help, and usage must be defined"
            logger.critical(message)
            raise NameError(message)

    def register(self, subparser: _SubParsersAction) -> None:  # pylint: disable=unused-argument
        """
        Implementation is required. Configure all subcommand options here. This
        method is called at application startup to register the subcommand's
        options with the main parser object.

        Args:
            subparser: the subparser object provided by the main method to register this subcommand with

        Returns:
            None
        """

        message = "The register method must be implemented"
        logger.critical(message)
        raise NameError(message)

    def exec(self, args: Namespace, extra: List[str]) -> None:  # pylint: disable=unused-argument
        """
        Implementation is required. Self-contained logic of the subcommand (ie. the subcommand's 'main' method).

        Args:
            args: default & user-supplied responses to all configured options
            extra: arguments provided by the user which may or may not correspond to configured options

        Returns:
            None
        """

        message = "The exec method must be implemented"
        logger.critical(message)
        raise NameError(message)



from flask import Blueprint, Response, request

from mmpm.api.constants import http
from mmpm.api.endpoints.endpoint import Endpoint
from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.database import MagicMirrorDatabase
from mmpm.magicmirror.magicmirror import MagicMirror
from mmpm.magicmirror.package import MagicMirrorPackage, RemotePackage

logger = MMPMLogFactory.get_logger(__name__)


class Packages(Endpoint):
    """
    Endpoint to handle installation, removal, upgrading, and updating of MagicMirrorPackages
    """

    def __init__(self):
        self.name = "packages"
        self.blueprint = Blueprint(self.name, __name__, url_prefix=f"/api/{self.name}")
        self.db = MagicMirrorDatabase()
        self.magicmirror = MagicMirror()

        @self.blueprint.route("/", methods=[http.GET])
        def retrieve() -> Response:
            """
            A Flask route method for retrieving a list of all MagicMirror packages.

            Parameters:
                None

            Returns:
                Response: A Flask Response object containing a list of all packages or an error message.
            """

            logger.info("Loading database")
            self.db.load()

            if not self.db.packages:
                message = "Failed to load database"
                logger.error(message)
                return self.failure(message)

            logger.info("Sending back current packages")

            # this section here is very inefficient, but honestly, there's not much of a way around
            # it if we want the user to have information about the status of the packages on the web app
            # also, the data isn't THAT large, so it's fine, but it's definitely kinda gross
            upgradable = [MagicMirrorPackage(**pkg) for pkg in self.db.upgradable()["packages"]]
            pkgs = self.db.packages

            for pkg in pkgs:
                if pkg in upgradable:
                    pkg.is_upgradable = True

            return self.success([package.serialize(full=True) for package in pkgs])

        @self.blueprint.route("/install", methods=[http.POST])
        def install() -> Response:
            """
            A Flask route method for installing selected MagicMirror packages.

            Parameters:
                None

            Returns:
                Response: A Flask Response object containing the status of each installation attempt.
            """

            packages = request.get_json()["packages"]
            success = []
            failure = []

            for package in packages:
                pkg = MagicMirrorPackage(**package)

                if pkg.install():
                    logger.debug(f"Installed {pkg.title}")
                    success.append(package)
                else:
                    logger.debug(f"Removing {pkg.title} due to installation failure. Please try reinstalling manually.")
                    failure.append(package)
                    pkg.remove()

            return self.success({"success": success, "failure": failure})

        @self.blueprint.route("/remove", methods=[http.POST])
        def remove() -> Response:
            """
            A Flask route method for removing selected MagicMirror packages.

            Parameters:
                None

            Returns:
                Response: A Flask Response object containing the status of each removal attempt.
            """

            packages = request.get_json()["packages"]
            success = []
            failure = []

            for package in packages:
                pkg = MagicMirrorPackage(**package)

                if pkg.remove():
                    logger.debug(f"Removed {pkg.title}")
                    success.append(package)
                else:  # honestly, this should never fail anyway
                    logger.debug(f"Failed to remove {pkg.title}")
                    failure.append(package)

            return self.success({"success": success, "failure": failure})

        @self.blueprint.route("/upgrade", methods=[http.POST])
        def upgrade() -> Response:
            """
            A Flask route method for upgrading selected MagicMirror packages.

            Parameters:
                None

            Returns:
                Response: A Flask Response object containing the status of each upgrade attempt.
            """

            packages = request.get_json()["packages"]
            success = []
            failure = []

            for package in packages:
                pkg = MagicMirrorPackage(**package)

                if pkg.upgrade():
                    logger.debug(f"Upgraded {pkg.title}")
                    success.append(package)
                else:
                    logger.debug(f"Failed to upgrade {pkg.title}")
                    failure.append(package)

            return self.success({"success": success, "failure": failure})

        @self.blueprint.route("/mm-pkg/add", methods=[http.POST])
        def add_mm_pkg() -> Response:
            """
            A Flask route method for adding a custom MagicMirror package.

            Parameters:
                None

            Returns:
                Response: A Flask Response object indicating the success or failure of adding the custom package.
            """

            package = MagicMirrorPackage(**request.get_json()["package"])

            if self.db.add_mm_pkg(package.title, package.author, package.repository, package.description):
                return self.success(f"Added custom package named {package.title}")

            return self.failure("Failed to add custom package. See logs for details.")

        @self.blueprint.route("/mm-pkg/remove", methods=[http.POST])
        def remove_mm_pkg() -> Response:
            """
            A Flask route method for removing custom MagicMirror packages.

            Parameters:
                None

            Returns:
                Response: A Flask Response object containing the status of each custom package removal attempt.
            """

            packages = request.get_json()["packages"]
            success = []
            failure = []

            for package in packages:
                pkg = MagicMirrorPackage(**package)

                if self.db.remove_mm_pkg(pkg.title):
                    success.append(package)
                else:
                    failure.append(package)

            return self.success({"success": success, "failure": failure})

        @self.blueprint.route("/details", methods=[http.POST])
        def details() -> Response:
            """
            A Flask route method for retrieving detailed information about a selected MagicMirror package.

            Parameters:
                None

            Returns:
                Response: A Flask Response object containing detailed information about the package or an error message.
            """

            package = request.get_json()["packages"][0]
            remote = RemotePackage(MagicMirrorPackage(**package))
            health = remote.health()

            for status in health.values():
                if status["error"]:
                    message = status["error"]
                    logger.error(message)
                    return self.failure(message, code=400)
                elif status["warning"]:
                    message = status["warning"]
                    logger.warning(message)
                    return self.failure(message, code=400)

            return self.success(remote.serialize())



import json

from flask import Blueprint, Response, request

from mmpm.api.constants import http
from mmpm.api.endpoints.endpoint import Endpoint
from mmpm.constants import paths
from mmpm.env import MMPM_DEFAULT_ENV, MMPMEnv
from mmpm.log.factory import MMPMLogFactory

logger = MMPMLogFactory.get_logger(__name__)


class Env(Endpoint):
    """
    A Flask endpoint class for handling operations related to the MMPM environment settings,
    including retrieving the current environment, the default environment, and updating the environment settings.
    """

    def __init__(self):
        self.name = "env"
        self.blueprint = Blueprint(self.name, __name__, url_prefix=f"/api/{self.name}")
        self.env = MMPMEnv()

        @self.blueprint.route("/", methods=[http.GET])
        def retrieve() -> Response:
            """
            A Flask route method for retrieving the current MMPM environment settings.

            Parameters:
                None

            Returns:
                Response: A Flask Response object containing the current MMPM environment settings.
            """

            logger.info("Sending back current MMPM Env")
            return self.success(self.env.get())

        @self.blueprint.route("/default", methods=[http.GET])
        def default() -> Response:
            """
            A Flask route method for retrieving the default MMPM environment settings.

            Parameters:
                None

            Returns:
                Response: A Flask Response object containing the default MMPM environment settings.
            """

            logger.info("Sending back default MMPM Env")
            return self.success({key: str(value) for key, value in MMPM_DEFAULT_ENV.items()})

        @self.blueprint.route("/update", methods=[http.POST])
        def update() -> Response:
            """
            A Flask route method for updating MMPM environment settings.

            Parameters:
                None

            Returns:
                Response: A Flask Response object indicating the success or failure of the update operation.
            """

            updated_env = request.get_json()["env"]

            with open(paths.MMPM_ENV_FILE, "w", encoding="utf-8") as env:
                try:
                    json.dump(updated_env, env, indent=2)
                except Exception as error:
                    message = f"Failed to updated env: {error}"
                    logger.error(message)
                    return self.failure(message)

            logger.info(f"Updating MMPM Env with {updated_env}")
            return self.success({"updated": True})



from flask import Blueprint, Response, jsonify

from mmpm.api.constants import http
from mmpm.api.endpoints.endpoint import Endpoint
from mmpm.log.factory import MMPMLogFactory

logger = MMPMLogFactory.get_logger(__name__)


class Index(Endpoint):
    """
    Endpoint that returns a formatted list of endpoints served by the MMPM API
    """

    def __init__(self, url_map):
        super().__init__()
        self.name = "index"
        self.blueprint = Blueprint(self.name, __name__, url_prefix="/api/index")
        self.url_map = url_map

        @self.blueprint.route("/", methods=[http.GET])
        def default() -> Response:
            """
            Returns a list of endpoints available in the API

            Parameters:
                None

            Returns:
                endpoints (List[Dict]): A list of objects that contain 'url' and 'methods' fields
            """
            logger.info("Sending back url map")
            rules = [(str(url), list(url.methods)) for url in self.url_map.iter_rules()]
            formatted_rules = [{"url": rule[0], "methods": rule[1]} for rule in rules]
            return jsonify(formatted_rules[1:])



"""Command line options for 'log' subcommand"""

from mmpm.log.factory import MMPMLogFactory
from mmpm.subcommands.sub_cmd import SubCmd

logger = MMPMLogFactory.get_logger(__name__)


class Logs(SubCmd):
    """
    The 'Logs' subcommand allows users to view all or specific log files and create an archive of them

    Custom Attributes:
        None
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "logs"
        self.help = f"Display, tail, or zip the {self.app_name} log files"
        self.usage = f"{self.app_name} {self.name} [--<option(s)>]"

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

        self.parser.add_argument(
            "-t",
            "--tail",
            action="store_true",
            help=f"Tail {self.app_name} log file(s)",
            dest="tail",
        )

        self.parser.add_argument(
            "-z",
            "--zip",
            action="store_true",
            help=f"Zip {self.app_name} CLI and/or UI log file(s)",
            dest="zip",
        )

    def exec(self, args, extra):
        if extra:
            logger.error(f"Extra arguments are not accepted. See '{self.app_name} {self.name} --help'")
        elif args.zip:
            MMPMLogFactory.archive()
        else:
            MMPMLogFactory.display(tail=args.tail)



"""Command line options for 'db' subcommand"""

import json

from pygments import highlight
from pygments.formatters.terminal import TerminalFormatter
from pygments.lexers.data import JsonLexer

from mmpm.constants import color
from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.database import MagicMirrorDatabase
from mmpm.subcommands.sub_cmd import SubCmd

logger = MMPMLogFactory.get_logger(__name__)


class Db(SubCmd):
    """
    The 'Db' subcommand allows users to refresh or display basic details about the MagicMirror packages database.

    Custom Attributes:
        database (MagicMirrorDatabase): An instance of the MagicMirrorDatabase class for managing the database.
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "db"
        self.help = "Display database metadata, or display raw database contents"
        self.usage = f"{self.app_name} {self.name} [--<option>]"
        self.database = MagicMirrorDatabase()

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

        group = self.parser.add_mutually_exclusive_group()

        group.add_argument(
            "-i",
            "--info",
            action="store_true",
            help="display database metadata",
            dest="info",
        )

        group.add_argument(
            "-d",
            "--dump",
            action="store_true",
            help="dump the database contents to stdout as JSON",
            dest="dump",
        )

    def exec(self, args, extra):
        if not self.database.is_initialized():
            self.database.load()

        if extra:
            logger.error(f"Extra arguments are not accepted. See '{self.app_name} {self.name} --help'")
            return

        if args.info:
            info = self.database.info()
            convert_string = lambda s: " ".join([s.split("_")[0].capitalize()] + [word.lower() for word in s.split("_")[1:]])

            for key, value in info.items():
                print(f"{color.n_green(convert_string(key))}:\n\t{value}\n")

        elif args.dump:
            print(
                highlight(
                    json.dumps(
                        self.database.packages,
                        indent=2,
                        default=lambda package: package.serialize(),
                    ),
                    JsonLexer(),
                    TerminalFormatter(),
                ),
            )
        else:
            logger.error(f"No arguments provided. See '{self.app_name} {self.name} --help'")



import os
import shutil
from pathlib import Path
from time import sleep
from typing import List

import socketio

from mmpm.constants import color
from mmpm.env import MMPMEnv
from mmpm.log.factory import MMPMLogFactory
from mmpm.singleton import Singleton
from mmpm.utils import kill_pids_of_process, run_cmd

logger = MMPMLogFactory.get_logger(__name__)


class MagicMirrorClientFactory:
    """
    A factory class for creating SocketIO clients to interact with MagicMirror modules. It
    facilitates communication between MMPM and MagicMirror via websockets.
    """

    @staticmethod
    def create_client(event: str, data: dict, namespace: str = "/MMM-mmpm"):
        """
        Creates and configures a SocketIO client for communicating with MagicMirror modules.

        Parameters:
            event (str): The event name to emit to the MagicMirror module.
            data (dict): The data payload to send with the event.
            namespace (str): The namespace for the SocketIO connection.

        Returns:
            socketio.Client: Configured SocketIO client, or None on failure.
        """

        client = None

        if not event:
            logger.error("No event name provided")
            return client

        try:
            client = socketio.Client(reconnection=True, request_timeout=300)
        except socketio.exceptions.SocketIOError as error:
            logger.fatal(f"Failed to create SocketIO client: {error}")

        @client.on("connect", namespace=namespace)
        def connect():
            logger.debug(f"Connected to MagicMirror websocket in {namespace}")
            client.emit(event, namespace=namespace, data=data)
            logger.debug(f"Emitting {data}")

        @client.event
        def connect_error(error):
            logger.error("Failed to connect to MagicMirror websocket. Is the MMPM_MAGICMIRROR_URI environment variable set properly?")
            logger.debug(f"Error when connecting to MagicMirror websocket: {error}")

        @client.on("disconnect", namespace=namespace)
        def disconnect():
            logger.debug("Disconnected from MagicMirror websocket")

        @client.on("ACTIVE_MODULES", namespace=namespace)
        def active_modules(data):
            logger.debug("Received active modules from MMPM MagicMirror module")

            if not data:
                logger.error("No data was received. Is the MMPM_MAGICMIRROR_URI environment variable set properly?")

            for module in [json_data for index, json_data in enumerate(data) if json_data not in data[index + 1 :]]:
                print(f"{color.n_green(module['name'])}\n  hidden: {'true' if module['hidden'] else 'false'}\n  key: {module['key'] + 1}\n")

            logger.debug("Disconnecting from client")
            client.disconnect()

        @client.on("MODULES_TOGGLED", namespace=namespace)
        def modules_toggled(data):
            logger.debug("Received toggled modules from MMPM MagicMirror module")

            if not data:
                logger.error("Unable to find provided module(s)")

            logger.debug("Disconnecting from client")
            client.disconnect()

        return client


class MagicMirrorController(Singleton):
    """
    Controller class for managing and interacting with the MagicMirror application. It provides
    methods to start, stop, restart, and check the status of MagicMirror, as well as to show
    or hide specific modules.
    """

    def __init__(self):
        self.env = MMPMEnv()
        self.factory = MagicMirrorClientFactory()

    def status(self):
        """
        Checks the status of MagicMirror by attempting to connect to it using a SocketIO client.

        Returns:
            True if the connection is successful, False otherwise.
        """

        try:
            client = self.factory.create_client("FROM_MMPM_APP_get_active_modules", {})
            client.connect(self.env.MMPM_MAGICMIRROR_URI.get())
            return True
        except (OSError, BrokenPipeError, Exception) as error:
            logger.error(f"Failed to connect to MagicMirror, closing socket. Is MagicMirror running? : {error}")

        return False

    def hide(self, modules: List[str]):
        """
        Hides specified modules in MagicMirror using a SocketIO client.

        Parameters:
            modules (List[str]): List of module names or identifiers to hide.

        Returns:
            True if the operation is successful, False otherwise.
        """

        try:
            client = self.factory.create_client(
                "FROM_MMPM_APP_toggle_modules",
                {"directive": "hide", "modules": modules},
            )

            client.connect(self.env.MMPM_MAGICMIRROR_URI.get())
            logger.info(f"Hid modules with keys: {', '.join(modules)}")
            return True
        except (OSError, BrokenPipeError, Exception) as error:
            logger.error(f"Failed to connect to MagicMirror, closing socket. Is MagicMirror running? : {error}")

        return False

    def show(self, modules: List[str]):
        """
        Hides specified modules in MagicMirror using a SocketIO client.

        Parameters:
            modules (List[str]): List of module names or identifiers to hide.

        Returns:
            True if the operation is successful, False otherwise.
        """

        try:
            client = self.factory.create_client(
                "FROM_MMPM_APP_toggle_modules",
                data={"directive": "show", "modules": modules},
            )

            client.connect(self.env.MMPM_MAGICMIRROR_URI.get())
            logger.info(f"Made modules visible with keys: {', '.join(modules)}")
            return True
        except (OSError, BrokenPipeError, Exception) as error:
            logger.error(f"Failed to connect to MagicMirror, closing socket. Is MagicMirror running? : {error}")

        return False

    def start(self) -> bool:
        """
        Launches MagicMirror using pm2, if found, otherwise a 'npm start' is run as
        a background process

        Parameters:
            None

        Returns:
            True if the operation is successful, False otherwise.
        """
        command = ["npm", "run", "start"]

        pm2_process: str = self.env.MMPM_MAGICMIRROR_PM2_PROCESS_NAME.get()
        compose_file: str = self.env.MMPM_MAGICMIRROR_DOCKER_COMPOSE_FILE.get()

        if compose_file:
            logger.debug(f"Docker compose file set as {compose_file}")
            command = ["docker", "compose", "-f", compose_file, "up", "-d"]
        elif pm2_process:
            logger.debug(f"pm2 process set as {pm2_process}")
            command = ["pm2", "start", pm2_process]

        if not shutil.which(command[0]):
            logger.error(f"{command[0]} not found in PATH. Unable to start MagicMirror")
            return False

        root = self.env.MMPM_MAGICMIRROR_ROOT.get()
        root.mkdir(exist_ok=True, parents=True)

        if not Path(root / "node_modules").exists():
            logger.error("MagicMirror dependencies have not been installed. Please run `mmpm mm-ctl --install` first.")
            return False

        os.chdir(root)

        logger.debug(f"Attempting to start MagicMirror using {' '.join(command)} ")
        error_code, _, stderr = run_cmd(
            command,
            message="Starting MagicMirror",
            background=bool(command[0] == "npm"),
        )

        if error_code:
            logger.error(stderr)
            return False

        logger.info("Started MagicMirror")
        return True

    def stop(self) -> bool:
        """
        Stops MagicMirror using pm2, if found, otherwise the associated
        processes are killed

        Parameters:
            None

        Returns:
            success (bool): True if successful, False if failure
        """

        command: List[str] = []

        pm2_process: str = self.env.MMPM_MAGICMIRROR_PM2_PROCESS_NAME.get()
        compose_file: str = self.env.MMPM_MAGICMIRROR_DOCKER_COMPOSE_FILE.get()

        if compose_file:
            logger.debug(f"docker-compose file set as {compose_file}")
            command = ["docker", "compose", "-f", compose_file, "stop"]
        elif pm2_process:
            logger.debug(f"pm2 process set as {pm2_process}")
            command = ["pm2", "stop", pm2_process]

        if command and shutil.which(command[0]):
            logger.debug(f"Attempting to stop MagicMirror using '{command[0]}'")
            # pm2 and docker-compose cause stdout/stderr to flip
            error_code, _, stderr = run_cmd(command, message="Stopping MagicMirror")

            if error_code:
                logger.error(stderr)
                return False

            logger.debug(f"Stopped MagicMirror using '{command[0]}'")
            logger.info("Stopped MagicMirror")
            return True

        logger.debug("Stopping electron processes associated with MagicMirror")
        kill_pids_of_process("electron")
        logger.info("Stopped MagicMirror")
        return True

    def restart(self) -> bool:
        """
        Restarts MagicMirror using pm2, if found, otherwise the associated
        processes are killed and 'npm start' is re-run a background process

        Parameters:
            None

        Returns:
            success (bool): True if successful, False if failure
        """
        if not self.stop():
            return False

        sleep(2)

        if not self.start():
            return False

        return True



"""Command line options for 'list' subcommand"""

from mmpm.constants import color
from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.database import MagicMirrorDatabase
from mmpm.magicmirror.package import MagicMirrorPackage
from mmpm.subcommands.sub_cmd import SubCmd

logger = MMPMLogFactory.get_logger(__name__)


class List(SubCmd):
    """
    The 'List' subcommand allows users to list items such as installed packages, available packages,
    available upgrades, package categories, and more.

    Custom Attributes:
        database (MagicMirrorDatabase): An instance of the MagicMirrorDatabase class for managing the database.
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "list"
        self.help = "List items such as installed packages, packages available, available upgrades, etc"
        self.usage = f"{self.app_name} {self.name} [--<option(s)>]"
        self.database = MagicMirrorDatabase()

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

        self.parser.add_argument(
            "-t",
            "--title-only",
            action="store_true",
            help="display the title only of packages (used with -c, -a, -e, or -i)",
            dest="title_only",
        )

        group = self.parser.add_mutually_exclusive_group()

        group.add_argument(
            "-a",
            "--all",
            action="store_true",
            help="list all available packages in the marketplace",
            dest="all",
        )

        group.add_argument(
            "-i",
            "--installed",
            action="store_true",
            help="list all locally installed packages",
            dest="installed",
        )

        group.add_argument(
            "-e",
            "--exclude-installed",
            action="store_true",
            help="list all available packages in the marketplace, excluding locally installed packages",
            dest="exclude_installed",
        )

        group.add_argument(
            "-c",
            "--categories",
            action="store_true",
            help="list all available package categories",
            dest="categories",
        )

        group.add_argument(
            "--upgradable",
            action="store_true",
            help="list packages that have available upgrades",
            dest="upgradable",
        )

    def exec(self, args, extra):
        if not self.database.is_initialized():
            self.database.load()

        if args.installed:
            for package in self.database.packages:
                if package.is_installed:
                    package.display(title_only=args.title_only, hide_installed_indicator=True)

        elif args.all or args.exclude_installed:
            for package in self.database.packages:
                package.display(title_only=args.title_only, exclude_installed=args.exclude_installed)

        elif args.categories:
            categories = {package.category for package in self.database.packages}

            if args.title_only:
                for category in categories:
                    print(category)
                return

            for category in categories:
                package_count = sum(1 for package in self.database.packages if package.category == category)
                print(color.n_green(category), f"\n\tPackages: {package_count}\n")

        elif args.upgradable:
            app_label: str = f"{color.n_cyan('application')}"
            pkg_label: str = f"{color.n_cyan('package')}"

            upgrades_available: bool = False
            upgradable = self.database.upgradable()

            if upgradable["packages"] or upgradable["mmpm"] or upgradable["MagicMirror"]:
                upgrades_available = True

            for package in upgradable["packages"]:
                print(color.n_green(MagicMirrorPackage(**package).title), f"[{pkg_label}]")

            if upgradable["mmpm"]:
                print(f"{color.n_green('mmpm')} [{app_label}]")

            if upgradable["MagicMirror"]:
                print(f"{color.n_green('MagicMirror')} [{app_label}]")

            if upgrades_available:
                print("Run `mmpm upgrade` to upgrade packages/applications")
            else:
                logger.info("No upgrades available")

        else:
            logger.error(f"No arguments provided. See '{self.app_name} {self.name} --help'")



"""Command line options for 'search' subcommand"""

from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.database import MagicMirrorDatabase
from mmpm.subcommands.sub_cmd import SubCmd

logger = MMPMLogFactory.get_logger(__name__)


class Search(SubCmd):
    """
    The 'Search' subcommand allows users to search for packages matching a name or within a category

    Custom Attributes:
        database (MagicMirrorDatabase): An instance of the MagicMirrorDatabase class for managing the database.
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "search"
        self.help = "Search for MagicMirror packages"
        self.usage = f"{self.app_name} {self.name} <query> [--<option(s)>]"
        self.database = MagicMirrorDatabase()

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

        self.parser.add_argument(
            "-t",
            "--title-only",
            action="store_true",
            help="only show the title of the packages matching the search results",
            dest="title_only",
        )

        self.parser.add_argument(
            "-c",
            "--case-sensitive",
            action="store_true",
            help="search for packages using a case-sensitive query",
            dest="case_sensitive",
        )

        self.parser.add_argument(
            "-e",
            "--exclude-installed",
            action="store_true",
            help="exclude installed packages from search results",
            dest="exclude_installed",
        )

    def exec(self, args, extra):
        if not extra:
            logger.error(f"No arguments provided. See '{self.app_name} {self.name} --help'")
            return

        if len(extra) > 1:
            logger.fatal(f"Too many arguments. `{self.app_name} {args.subcmd}` only accepts one search argument")
            return

        if not self.database.is_initialized():
            self.database.load()

        results = self.database.search(extra[0], case_sensitive=args.case_sensitive, title_only=args.title_only)

        if not results:
            logger.error(f"No results found for '{extra[0]}'")

        for package in results:
            package.display(title_only=args.title_only)



import os
import shutil
import sys
from os import chdir
from pathlib import Path, PosixPath
from typing import Optional

from mmpm.constants import color, urls
from mmpm.env import MMPMEnv
from mmpm.log.factory import MMPMLogFactory
from mmpm.singleton import Singleton
from mmpm.utils import repo_up_to_date, run_cmd

logger = MMPMLogFactory.get_logger(__name__)


class MagicMirrorConfigs(Singleton):
    def __init__(self) -> None:
        self.env = MMPMEnv()

    def get(self, name: str) -> Optional[Path]:
        if name == self.config_js.name:
            return self.config_js
        elif name == self.config_js_sample.name:
            return self.config_js_sample
        elif name == self.custom_css.name:
            return self.custom_css
        elif name == self.custom_css_sample.name:
            return self.custom_css_sample

        logger.error(f"File named {name} doesn't exist within MagicMirrorConfigs")
        return None

    @property
    def config_js(self) -> Path:
        root: Path = self.env.MMPM_MAGICMIRROR_ROOT.get()
        return root / "config" / "config.js"

    @property
    def custom_css(self) -> Path:
        root: Path = self.env.MMPM_MAGICMIRROR_ROOT.get()
        legacy_css = root / "config" / "custom.css"
        return legacy_css if legacy_css.exists() else Path(root / "css" / "custom.css")

    @property
    def custom_css_sample(self) -> Path:
        root: Path = self.env.MMPM_MAGICMIRROR_ROOT.get()
        legacy_css_sample = root / "config" / "custom.css.sample"
        return legacy_css_sample if legacy_css_sample.exists() else root / "css" / "custom.css.sample"

    @property
    def config_js_sample(self) -> Path:
        root: Path = self.env.MMPM_MAGICMIRROR_ROOT.get()
        return root / "config" / "config.js.sample"


class MagicMirror(Singleton):
    """
    A class for managing the MagicMirror application including update, upgrade, install,
    and removal functionalities.
    """

    def __init__(self):
        self.env = MMPMEnv()

    def update(self):
        """
        Checks for updates available to the MagicMirror repository

        Parameters:
            None

        Returns:
            bool: True if an upgrade is available, False otherwise
        """
        magicmirror_root: PosixPath = self.env.MMPM_MAGICMIRROR_ROOT.get()

        if not magicmirror_root.exists() or not (magicmirror_root / ".git").exists():
            logger.warning("MagicMirror application directory not found or not a git repo.")
            return False

        logger.debug("Checking to see if MagicMirror is up to date")

        chdir(magicmirror_root)
        print(f"Retrieving: {urls.MAGICMIRROR_REPO_URL} [{color.n_cyan('MagicMirror')}]")

        try:
            can_upgrade = repo_up_to_date(magicmirror_root)
        except KeyboardInterrupt:
            logger.info("User killed process with CTRL-C")
            sys.exit(127)

        return can_upgrade

    def upgrade(self):
        """
        Handles upgrade processs of MagicMirror by pulling changes from MagicMirror
        repo, and installing dependencies.

        Parameters:
            None

        Returns:
            success (bool): True if successful else False
        """
        print(f"Upgrading {color.n_green('MagicMirror')}")

        root = self.env.MMPM_MAGICMIRROR_ROOT
        root_dir: PosixPath = root.get()

        if not root_dir.exists():
            message = f"{root.name}='{root_dir}' does not exist. Is the {root.name} set properly?"
            logger.error(message)
            return False

        os.chdir(root_dir)

        error_code, _, stderr = run_cmd(["git", "checkout", "."], progress=False)

        if error_code:
            message = "Failed to checkout MagicMirror repo for clean upgrade"
            logger.error(f"{message}. See `mmpm logs` for details")
            return stderr

        error_code, _, stderr = run_cmd(["git", "pull"], progress=False)

        if error_code:
            message = "Failed to upgrade MagicMirror"
            logger.error(f"{message}. See `mmpm logs` for details")
            return stderr

        error_code, _, stderr = run_cmd(["npm", "install"], progress=True)

        if error_code:
            logger.error(stderr)
            return False

        print("Upgrade complete! Restart MagicMirror for the changes to take effect")
        return True

    def install(self):
        """
        Installs MagicMirror by downloading the git repo and using NPM to install dependencies.

        Parameters:
            None

        Returns:
            bool: True upon success, False upon failure
        """

        root = self.env.MMPM_MAGICMIRROR_ROOT
        root_path: PosixPath = root.get()

        if root_path.exists() and Path(root_path / "modules").exists() and Path(root_path / "node_modules").exists():
            message = f"MagicMirror appears to already be installed in {root_path}. To install MagicMirror elsewhere, modify the {root.name} using 'mmpm open --env'"
            logger.fatal(message)
            return False

        for cmd in ["git", "npm"]:
            if not shutil.which(cmd):
                logger.fatal(f"'{cmd}' command not found. Please install '{cmd}', then re-run 'mmpm mm-ctl --install'")
                return False

        if not root_path.exists():
            root_path.mkdir(exist_ok=True)
            os.chdir(root_path.parent)

            error_code, _, stderr = run_cmd(
                ["git", "clone", urls.MAGICMIRROR_REPO_URL],
                progress=True,
                message="Downloading MagicMirror",
            )

            if error_code:
                logger.error(f"Failed to download MagicMirror: {stderr}")
                return False

        os.chdir(root_path)

        error_code, _, stderr = run_cmd(
            ["npm", "run", "install-mm"],
            progress=True,
            message="Installing MagicMirror",
        )

        if error_code:
            logger.error(f"Failed to install MagicMirror dependencies: {stderr}")
            return False

        logger.info("Installed MagicMirror!")
        print(f"Run {color.n_green('`mmpm mm-ctl --start`')} to start MagicMirror")
        return True

    def remove(self) -> bool:
        """
        Removes the MagicMirror installation from the specified directory.

        Parameters:
            None

        Returns:
            bool: True if the removal is successful, False otherwise.
        """

        root = self.env.MMPM_MAGICMIRROR_ROOT
        root_path: PosixPath = root.get()

        if not root_path.exists():
            message = f"'{root.name}'={root_path} does not exist. Is {root.name} set properly?"
            logger.fatal(message)
            return False

        if os.getcwd() == f"{root_path}":
            os.chdir("/tmp")

        shutil.rmtree(root_path, ignore_errors=True)
        logger.info("MagicMirror has been removed.")
        return True



"""Command line options for 'show' subcommand"""

from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.database import MagicMirrorDatabase
from mmpm.magicmirror.package import RemotePackage
from mmpm.subcommands.sub_cmd import SubCmd

logger = MMPMLogFactory.get_logger(__name__)


class Show(SubCmd):
    """
    The 'Show' subcommand allows users to view more details about a given MagicMirror package

    Custom Attributes:
        database (MagicMirrorDatabase): An instance of the MagicMirrorDatabase class for managing the database.
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "show"
        self.help = "Show details about one or more packages"
        self.usage = f"{self.app_name} {self.name} [--remote]"
        self.database = MagicMirrorDatabase()

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

        self.parser.add_argument(
            "-r",
            "--remote",
            action="store_true",
            help="display remote detail for package(s) from GitHub/GitLab/Bitbucket APIs",
            dest="remote",
        )

    def exec(self, args, extra):
        if not extra:
            logger.error(f"No arguments provided. See '{self.app_name} {self.name} --help'")

        if not self.database.is_initialized():
            self.database.load()

        if args.remote:
            health = RemotePackage.health()

            for status in health.values():
                if status["error"]:
                    logger.fatal(status["error"])
                elif status["warning"]:
                    logger.warning(status["warning"])

        for query in extra:
            results = self.database.search(query, title_only=True)

            if not results:
                logger.error(f"No results found for '{query}'")

            for package in results:
                logger.debug(f"Showing information for {package}")
                package.display(remote=args.remote, detailed=True)



from pathlib import Path

HOME_DIR = Path.home()
PM2_LOG_DIR = HOME_DIR / ".pm2" / "logs"
MMPM_CONFIG_DIR = HOME_DIR / ".config" / "mmpm"
MMPM_LOG_DIR = MMPM_CONFIG_DIR / "log"
MMPM_CLI_LOG_FILE = MMPM_LOG_DIR / "mmpm-cli.log"
MMPM_ENV_FILE = MMPM_CONFIG_DIR / "mmpm-env.json"
MMPM_CUSTOM_PACKAGES_FILE = MMPM_CONFIG_DIR / "mmpm-custom-packages.json"
MMPM_AVAILABLE_UPGRADES_FILE = MMPM_CONFIG_DIR / "mmpm-available-upgrades.json"
MAGICMIRROR_3RD_PARTY_PACKAGES_DB_FILE = MMPM_CONFIG_DIR / "MagicMirror-3rd-party-packages-db.json"
MAGICMIRROR_3RD_PARTY_PACKAGES_DB_LAST_UPDATE_FILE = MMPM_CONFIG_DIR / "MagicMirror-3rd-party-packages-db-last-update.json"

# Setup the directories and files
MMPM_CONFIG_DIR.mkdir(exist_ok=True, parents=True)
MMPM_LOG_DIR.mkdir(exist_ok=True)
MMPM_CLI_LOG_FILE.touch(exist_ok=True)
MMPM_ENV_FILE.touch(exist_ok=True)
MMPM_CUSTOM_PACKAGES_FILE.touch(exist_ok=True)
MMPM_AVAILABLE_UPGRADES_FILE.touch(exist_ok=True)
MAGICMIRROR_3RD_PARTY_PACKAGES_DB_FILE.touch(exist_ok=True)
MAGICMIRROR_3RD_PARTY_PACKAGES_DB_LAST_UPDATE_FILE.touch(exist_ok=True)



import importlib.resources as pkg_resources
import json
import os
from shutil import rmtree, which
from sys import executable

from mmpm.__version__ import version
from mmpm.constants import paths, urls
from mmpm.log.factory import MMPMLogFactory
from mmpm.singleton import Singleton
from mmpm.utils import run_cmd

logger = MMPMLogFactory.get_logger(__name__)


class MMPMui(Singleton):
    """
    Class responsible for managing the MMPM user interface. It provides methods to control and monitor
    the MMPM UI application, including starting, stopping, installing, and removing the UI.
    """

    def __init__(self):
        python = executable
        gunicorn = which("gunicorn") or f"{python} -m gunicorn"
        namespace = "mmpm"

        self.pm2_config_path = paths.MMPM_CONFIG_DIR / "ecosystem.json"

        self.pm2_ecosystem_config = {
            "apps": [
                {
                    "namespace": namespace,
                    "name": f"{namespace}.api",
                    "script": f"{gunicorn} -k gevent -b 0.0.0.0:{urls.MMPM_API_SERVER_PORT} mmpm.wsgi:app",
                    "version": version,
                    "watch": True,
                },
                {
                    "namespace": namespace,
                    "name": f"{namespace}.log-server",
                    "script": f"{gunicorn} -k geventwebsocket.gunicorn.workers.GeventWebSocketWorker -w 1 'mmpm.log.server:create()' -b 0.0.0.0:{urls.MMPM_LOG_SERVER_PORT}",
                    "version": version,
                    "watch": True,
                },
                {
                    "namespace": namespace,
                    "name": f"{namespace}.repeater",
                    "script": f"{gunicorn} -k geventwebsocket.gunicorn.workers.GeventWebSocketWorker -w 1 'mmpm.api.repeater:create()' -b 0.0.0.0:{urls.MMPM_REPEATER_SERVER_PORT}",
                    "version": version,
                    "watch": True,
                },
                {
                    "namespace": namespace,
                    "name": f"{namespace}.ui",
                    "script": f"{python} -m http.server -d {pkg_resources.files('mmpm').resolve() / 'ui'} -b 0.0.0.0 {urls.MMPM_UI_PORT}",
                    "version": version,
                    "watch": True,
                },
            ]
        }

    def create_pm2_config(self):
        """
        Creates or overwrites the PM2 configuration file for MMPM. This
        configuration file is used to manage the MMPM UI processes.

        Parameters:
            None

        Returns:
            None
        """

        logger.debug(f"Creating {self.pm2_config_path} file")

        self.pm2_config_path.parent.mkdir(exist_ok=True)
        self.pm2_config_path.touch(exist_ok=True)

        with open(self.pm2_config_path, mode="w", encoding="utf-8") as config:
            logger.debug(f"Writing PM2 Configuration to {self.pm2_config_path}")
            json.dump(self.pm2_ecosystem_config, config)

    def stop(self):
        """
        Stops the MMPM UI application using PM2.

        Parameters:
            None

        Returns:
            A tuple containing the process exit code, stdout, and stderr.
        """

        return run_cmd(["pm2", "stop", f"{self.pm2_config_path}"], message="Stopping MMPM UI")

    def delete(self):
        """
        Deletes the MMPM UI application from PM2's process list.

        Parameters:
            None

        Returns:
            A tuple containing the process exit code, stdout, and stderr.
        """

        return run_cmd(["pm2", "delete", f"{self.pm2_config_path}"], message="Removing MMPM UI")

    def start(self):
        """
        Starts the MMPM UI application using PM2.

        Parameters:
            None

        Returns:
            A tuple containing the process exit code, stdout, and stderr.
        """

        return run_cmd(["pm2", "start", f"{self.pm2_config_path}"], message="Installing MMPM UI")

    def install(self) -> bool:
        """
        Installs the MMPM UI by setting up necessary configurations and services. It creates the PM2
        configuration file and starts the MMPM UI application. If PM2 is not installed, logs a fatal error.

        Parameters:
            None

        Returns:
            True if the installation is successful, False otherwise.
        """

        if not which("pm2"):
            logger.fatal("pm2 is not in your PATH. Please run `npm install -g pm2`, and run the UI installation again.")
            return False

        self.create_pm2_config()
        error_code, _, stderr = self.start()

        if error_code:
            logger.error(f"Failed to install MMPM UI: {stderr}")
            return False

        return True

    def status(self) -> None:
        """
        Displays the current status of the MMPM UI processes using PM2.

        Parameters:
            None

        Returns:
            None
        """
        self.create_pm2_config()
        os.system("pm2 list mmpm")

    def remove(self) -> bool:
        """
        Removes the MMPM UI by stopping and deleting its processes from PM2 and cleaning up configuration
        files. Logs a fatal error if PM2 is not installed.

        Returns:
            True if the removal is successful, False otherwise.
        """
        if not which("pm2"):
            logger.fatal("pm2 is not in your PATH. Please run `npm install -g pm2`, and run the UI installation again.")
            return False

        self.create_pm2_config()
        error_code, _, stderr = self.delete()

        if error_code:
            logger.error(stderr)
            return False

        rmtree(self.pm2_config_path.parent, ignore_errors=True)
        return True



class __SingletonMeta(type):
    """
    This metaclass ensures that derived classes only create a single instance.
    The same instance is returned every time the class is instantiated.

    Attributes:
        _instances (dict): A dictionary mapping each class to its created instance.

    Methods:
        __call__(*args, **kwargs): Returns a single instance of the class, creating it if necessary.
    """

    _instances = {}  # type: ignore

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class Singleton(metaclass=__SingletonMeta):
    """
    Base class for implementing the Singleton design pattern.

    Classes derived from this base class will be Singletons, ensuring only one
    instance of the class is created throughout the program's lifecycle.

    Inherit from this class to create a Singleton.
    """



"""Command line options for 'remove' subcommand"""

from typing import List

from mmpm.constants import color
from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.database import MagicMirrorDatabase
from mmpm.subcommands.sub_cmd import SubCmd
from mmpm.utils import confirm

logger = MMPMLogFactory.get_logger(__name__)


class Remove(SubCmd):
    """
    The 'Remove' subcommand allows users to uninstall MagicMirror packages from their modules folder

    Custom Attributes:
        database (MagicMirrorDatabase): An instance of the MagicMirrorDatabase class for managing the database.
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "remove"
        self.help = "Remove installed MagicMirror packages"
        self.usage = f"{self.app_name} {self.name} <package(s)> [--yes]"
        self.database = MagicMirrorDatabase()

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

        self.parser.add_argument(
            "-y",
            "--yes",
            action="store_true",
            default=False,
            help="assume yes for user response and do not show prompt",
            dest="assume_yes",
        )

    def exec(self, args, extra):
        if not extra:
            logger.error(f"No arguments provided. See '{self.app_name} {self.name} --help'")
            return

        if not self.database.is_initialized():
            self.database.load()

        package_titles: List[str] = {package.title: package for package in self.database.packages}

        for name in extra:
            package = package_titles.get(name)

            if package is None:
                logger.error(f"'{name}' is not found in the installed packages")
                continue

            if not package.is_installed:
                logger.error(f"'{package.title}' is not installed")
                continue

            if not args.assume_yes and not confirm(f"Remove {package.title} ({package.repository})?"):
                continue

            if package.remove():
                logger.info(f"Removed {color.n_green(package.title)} ({package.repository})")



from colorama import Fore, Style

RESET = Style.RESET_ALL  # resets back to the original color of the terminal text

N_RED = Style.NORMAL + Fore.RED  # normal red
B_RED = Style.BRIGHT + Fore.RED  # bright red

N_MAGENTA = Style.NORMAL + Fore.MAGENTA  # normal magenta
B_MAGENTA = Style.BRIGHT + Fore.MAGENTA  # bright magenta

N_CYAN = Style.NORMAL + Fore.CYAN  # normal cyan
B_CYAN = Style.BRIGHT + Fore.CYAN  # bright cyan

N_YELLOW = Style.NORMAL + Fore.YELLOW  # normal yellow
B_YELLOW = Style.BRIGHT + Fore.YELLOW  # bright yellow

N_WHITE = Style.NORMAL + Fore.WHITE  # normal white
B_WHITE = Style.BRIGHT + Fore.WHITE  # bright white

N_GREEN = Style.NORMAL + Fore.GREEN  # normal green
B_GREEN = Style.BRIGHT + Fore.GREEN  # bright green

__reset = lambda text: str(text + RESET)
n_green = lambda text: N_GREEN + __reset(text)
n_cyan = lambda text: N_CYAN + __reset(text)
n_red = lambda text: N_RED + __reset(text)
n_magenta = lambda text: N_MAGENTA + __reset(text)
n_yellow = lambda text: N_YELLOW + __reset(text)

b_green = lambda text: B_GREEN + __reset(text)
b_cyan = lambda text: B_CYAN + __reset(text)
b_red = lambda text: B_RED + __reset(text)
b_magenta = lambda text: B_MAGENTA + __reset(text)
b_yellow = lambda text: B_YELLOW + __reset(text)



from datetime import datetime
from os import chdir
from shutil import make_archive

from flask import Blueprint, Response, send_file

from mmpm.api.constants import http
from mmpm.api.endpoints.endpoint import Endpoint
from mmpm.constants import paths
from mmpm.log.factory import MMPMLogFactory

logger = MMPMLogFactory.get_logger(__name__)


class Logs(Endpoint):
    """
    A Flask endpoint for handling requests related to log files.
    """

    def __init__(self):
        """
        Initializes the Logs endpoint with its name, blueprint, and route handler.
        """
        super().__init__()
        self.name = "logs"
        self.blueprint = Blueprint(self.name, __name__, url_prefix=f"/api/{self.name}")

        @self.blueprint.route("/archive", methods=[http.GET])
        def archive() -> Response:
            """
            Handles a GET request to create and serve a ZIP archive of log files.

            Returns:
                Response: A Flask Response object that enables the client to download the ZIP archive.
            """
            logger.debug("Creating zip of log files")
            chdir("/tmp")
            today = datetime.now()
            zip_file_name = f"mmpm-logs-{today.year}-{today.month}-{today.day}"
            logger.debug(f"Creating zip of log files named '{zip_file_name}'")
            archive_name = make_archive(zip_file_name, "zip", paths.MMPM_LOG_DIR)
            logger.debug(f"Archive created: {bool(archive_name)}")
            return send_file(f"/tmp/{zip_file_name}.zip", f"{zip_file_name}.zip", as_attachment=True)



"""Command line options for 'open' subcommand"""

from os import getenv, system
from pathlib import Path
from shutil import copyfile

from mmpm.constants import paths, urls
from mmpm.env import MMPMEnv
from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.magicmirror import MagicMirrorConfigs
from mmpm.subcommands.sub_cmd import SubCmd
from mmpm.ui import MMPMui
from mmpm.utils import run_cmd

logger = MMPMLogFactory.get_logger(__name__)


class Open(SubCmd):
    """
    The 'Open' subcommand allows users to open various configuration files, documentation,
    wikis, and the MagicMirror application itself.

    Custom Attributes:
        env (MMPMEnv): An instance of the MMPMEnv class for managing environment variables.
        ui (MMPMui): An instance of the MMPMui class for managing the MMPM UI.
        [Method] edit(self, file: PosixPath): see method docs
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "open"
        self.help = "Open config files, documentation, wikis, and MagicMirror itself"
        self.usage = f"{self.app_name} {self.name} [--<option>]"
        self.env = MMPMEnv()
        self.ui = MMPMui()
        self.mm_configs = MagicMirrorConfigs()

    def edit(self, file: Path) -> None:
        """
        Checks if the requested file exists, and if not, the file is created.
        Then, opens the file for editing using the system's default editor.

        Parameters:
            file (PosixPath): The file path to open for editing.

        Returns:
            None
        """

        if not file.exists():
            try:
                logger.warning(f"{file} does not exist. Creating file.")
                file.parent.mkdir(parents=True, exist_ok=True)
                file.touch(mode=0o664, exist_ok=True)
            except OSError as error:
                logger.fatal(f"Unable to create {file}: {str(error)}")

        logger.info(f"Opening {file} for user to edit")
        command = getenv("EDITOR", getenv("VISUAL", "edit"))
        system(f"{command} {file}")

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

        group = self.parser.add_mutually_exclusive_group()

        group.add_argument(
            "--config",
            action="store_true",
            help="open MagicMirror config/config.js file in your $EDITOR",
            dest="config",
        )

        group.add_argument(
            "--css",
            action="store_true",
            help="open MagicMirror css/custom.css file (if it exists) in your $EDITOR",
            dest="custom_css",
        )

        group.add_argument(
            "--ui",
            action="store_true",
            help="open the MMPM UI in your default browser",
            dest="ui",
        )

        group.add_argument(
            "--magicmirror",
            action="store_true",
            help="open MagicMirror in your default browser (uses the MMPM_MAGICMIRROR_URI address)",
            dest="magicmirror",
        )

        group.add_argument(
            "--mm-wiki",
            action="store_true",
            help="open the MagicMirror GitHub wiki in your default browser",
            dest="mm_wiki",
        )

        group.add_argument(
            "--mm-docs",
            action="store_true",
            help="open the MagicMirror documentation in your default browser",
            dest="mm_docs",
        )

        group.add_argument(
            "--mmpm-wiki",
            action="store_true",
            help="open the MMPM GitHub wiki in your default browser",
            dest="mmpm_wiki",
        )

        group.add_argument(
            "--env",
            action="store_true",
            help="open the MMPM run-time environment variables JSON configuration file in your $EDITOR",
            dest="mmpm_env",
        )

    def exec(self, args, extra):
        if extra:
            logger.error(f"Extra arguments are not accepted. See '{self.app_name} {self.name} --help'")
        elif args.config:
            config_js = self.mm_configs.config_js
            config_js_sample = self.mm_configs.config_js_sample

            if not config_js.stat().st_size and config_js_sample.exists():
                copyfile(config_js_sample, config_js)

            self.edit(config_js)

        elif args.custom_css:
            self.edit(self.mm_configs.custom_css)
        elif args.magicmirror:
            run_cmd(["xdg-open", self.env.MMPM_MAGICMIRROR_URI.get()], background=True)
        elif args.ui:
            run_cmd(["xdg-open", f"http://{urls.HOST}:{urls.MMPM_UI_PORT}"], background=True)
        elif args.mm_wiki:
            run_cmd(["xdg-open", urls.MAGICMIRROR_WIKI_URL], background=True)
        elif args.mm_docs:
            run_cmd(["xdg-open", urls.MAGICMIRROR_DOCUMENTATION_URL], background=True)
        elif args.mmpm_wiki:
            run_cmd(["xdg-open", urls.MMPM_WIKI_URL], background=True)
        elif args.mmpm_env:
            self.edit(paths.MMPM_ENV_FILE)
        else:
            logger.error(f"No arguments provided. See '{self.app_name} {self.name} --help'")



"""Command line options for 'update' subcommand"""

import mmpm.utils
from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.database import MagicMirrorDatabase
from mmpm.magicmirror.magicmirror import MagicMirror
from mmpm.subcommands.sub_cmd import SubCmd

logger = MMPMLogFactory.get_logger(__name__)


class Update(SubCmd):
    """
    The 'Update' subcommand checks for available updates on all installed packages, MagicMirror and MMPM itself

    Custom Attributes:
        database (MagicMirrorDatabase): An instance of the MagicMirrorDatabase class for managing the database.
        magicmirror (MagicMirror): An instance of the MagicMirror object (similar to a MagicMirrorPackage)
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "update"
        self.help = "Check for updates for installed packages, MMPM, and MagicMirror"
        self.usage = f"{self.app_name} {self.name}"
        self.magicmirror = MagicMirror()
        self.database = MagicMirrorDatabase()

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

    def exec(self, args, extra):
        if extra:
            logger.error(f"Extra arguments are not accepted. See '{self.app_name} {self.name} --help'")
            return

        self.database.load(update=True)

        can_upgrade_mmpm = mmpm.utils.update_available()
        can_upgrade_magicmirror = self.magicmirror.update()
        available_upgrades = self.database.update(
            can_upgrade_mmpm=can_upgrade_mmpm,
            can_upgrade_magicmirror=can_upgrade_magicmirror,
        )

        if not available_upgrades:
            print("Everything is up to date.")
            return

        print(f"{available_upgrades} upgrade(s) available. Run `mmpm list --upgradable` for details")



"""Command line options for 'db' subcommand"""

from shutil import which
from time import sleep

from mmpm.constants import urls
from mmpm.env import MMPMEnv
from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.database import MagicMirrorDatabase
from mmpm.subcommands.sub_cmd import SubCmd
from mmpm.ui import MMPMui
from mmpm.utils import confirm

logger = MMPMLogFactory.get_logger(__name__)


class Ui(SubCmd):
    """
    The 'Ui' subcommand allows users to interact with the MMPM user interface (UI), which includes
    displaying the UI URL, checking its status, installing, or removing it.

    Custom Attributes:
        database (MagicMirrorDatabase): An instance of the MagicMirrorDatabase class for managing the database.
        ui (MMPMui): An instance of the MMPMui class for managing the MMPM user interface.
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "ui"
        self.help = f"Interact with the {self.app_name} UI "
        self.usage = f"{self.app_name} {self.name} [--url] [--status] <install/remove>"
        self.database = MagicMirrorDatabase()
        self.ui = MMPMui()
        self.env = MMPMEnv()

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

        self.parser.add_argument(
            "-u",
            "--url",
            action="store_true",
            help=f"display the url of the {self.app_name} {self.name}",
            dest="url",
        )

        self.parser.add_argument(
            "--status",
            action="store_true",
            help=f"display the status of the {self.app_name} {self.name}",
            dest="status",
        )

        self.parser.add_argument(
            "--start",
            action="store_true",
            help=f"Start the {self.app_name} {self.name}",
            dest="start",
        )

        self.parser.add_argument(
            "--restart",
            action="store_true",
            help=f"Restart the {self.app_name} {self.name}",
            dest="restart",
        )

        self.parser.add_argument(
            "--stop",
            action="store_true",
            help=f"Stop the {self.app_name} {self.name}",
            dest="stop",
        )

        subparsers = self.parser.add_subparsers(
            dest="command",
            description=f"use `{self.app_name} {self.name} <subcommand> --help` to see more details",
            title=f"{self.app_name} {self.name} subcommands",
            metavar="",
        )

        install_parser = subparsers.add_parser("install", help=f"Install the {self.app_name} {self.name}")

        install_parser.add_argument(
            "-y",
            "--yes",
            action="store_true",
            help="assume yes",
            dest="assume_yes",
        )

        remove_parser = subparsers.add_parser("remove", help=f"Remove the {self.app_name} {self.name}")

        remove_parser.add_argument(
            "-y",
            "--yes",
            action="store_true",
            help="assume yes",
            dest="assume_yes",
        )

        reinstall_parser = subparsers.add_parser("reinstall", help=f"Reinstall the {self.app_name} {self.name}")

        reinstall_parser.add_argument(
            "-y",
            "--yes",
            action="store_true",
            help="assume yes",
            dest="assume_yes",
        )

    def exec(self, args, extra):
        if not which("pm2"):
            logger.fatal("pm2 is not in your PATH. Please run `npm install -g pm2`, and run the UI installation again.")
            return

        if not self.database.is_initialized():
            self.database.load()

        if self.env.MMPM_IS_DOCKER_IMAGE.get():
            logger.error("MMPM is a Docker image. Unable to interact with MMPM UI from CLI.")
            return

        if extra:
            logger.error(f"Extra arguments are not accepted. See '{self.app_name} {self.name} --help'")
            return

        if args.url:
            print(f"http://{urls.HOST}:{urls.MMPM_UI_PORT}")

        elif args.status:
            self.ui.status()

        elif args.start:
            self.ui.start()

        elif args.stop:
            self.ui.stop()

        elif args.restart:
            self.ui.stop()
            sleep(1)
            self.ui.start()

        elif args.command == "install":
            if not args.assume_yes and not confirm("Are you sure you want to install the MMPM UI?"):
                return

            if not self.ui.install():
                logger.error("Failed to install MMPM UI")
                self.ui.delete()
            else:
                logger.info("Installed MMPM-UI")
                print("Run `mmpm ui --url` to display the UI address, or execute `mmpm open --ui` to open it.")

        elif args.command == "reinstall":
            if not args.assume_yes and not confirm("Are you sure you want to reinstall the MMPM UI?"):
                return

            if not self.ui.remove():
                logger.error("Failed to remove MMPM UI")
                self.ui.delete()

            sleep(2)

            if not self.ui.install():
                logger.error("Failed to install MMPM UI")
                self.ui.delete()
            else:
                logger.info("Reinstalled MMPM-UI")
                print("Run `mmpm ui --url` to display the UI address, or execute `mmpm open --ui` to open it.")

        elif args.command == "remove":
            if not args.assume_yes and not confirm("Are you sure you want to remove the MMPM UI?"):
                return

            if not self.ui.remove():
                logger.error("Failed to remove MMPM UI")
                self.ui.delete()
        else:
            logger.error(f"No arguments provided. See '{self.app_name} {self.name} --help'")



"""Command line options for 'version' subcommand"""

from mmpm.__version__ import version
from mmpm.log.factory import MMPMLogFactory
from mmpm.subcommands.sub_cmd import SubCmd

logger = MMPMLogFactory.get_logger(__name__)


class Version(SubCmd):
    """
    The 'Version' subcommand displays the current version number of MMPM

    Custom Attributes:
        None
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "version"
        self.help = f"Display {self.app_name} application version"
        self.usage = f"{self.app_name} {self.name}"

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

    def exec(self, args, extra):
        if extra:
            logger.error(f"{self.name} does not take arguments")
        else:
            print(version)



from flask import Blueprint, Response

import mmpm.__version__
import mmpm.utils
from mmpm.api.constants import http
from mmpm.api.endpoints.endpoint import Endpoint
from mmpm.env import MMPMEnv
from mmpm.log.factory import MMPMLogFactory

logger = MMPMLogFactory.get_logger(__name__)


class Mmpm(Endpoint):
    """
    A Flask endpoint class for interacting with the MMPM application in more 'meta' manner.
    """

    def __init__(self):
        self.name = "mmpm"
        self.blueprint = Blueprint(self.name, __name__, url_prefix=f"/api/{self.name}")
        self.env = MMPMEnv()

        @self.blueprint.route("/version", methods=[http.GET])
        def version() -> Response:
            """
            A Flask route method for retrieving the current MMPM version.

            Parameters:
                None

            Returns:
                Response: A Flask Response object containing the current MMPM version.
            """
            return self.success(mmpm.__version__.version)

        @self.blueprint.route("/upgrade", methods=[http.GET])
        def upgrade() -> Response:
            """
            A Flask route method for upgrading MMPM.

            Parameters:
                None

            Returns:
                Response: A Flask Response object indicating success or failure of the upgrade operation.
            """
            if self.env.MMPM_IS_DOCKER_IMAGE.get():
                self.failure("MMPM is a Docker image. MMPM must be upgraded using the Docker CLI.")
            if mmpm.utils.upgrade():
                return self.success("Upgrade MMPM")

            return self.failure("Failed to upgrade MMPM. See logs for details.")



"""Command line options for 'env' subcommand"""

from mmpm.env import MMPMEnv
from mmpm.log.factory import MMPMLogFactory
from mmpm.subcommands.sub_cmd import SubCmd

logger = MMPMLogFactory.get_logger(__name__)


class Env(SubCmd):
    """
    The 'Env' subcommand allows users to view the current environment variables set for MMPM

    Custom Attributes:
        env (MMPMEnv): A singleton of MMPMEnv which contains environment variables
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "env"
        self.help = f"Display the {self.name} environment variables and their value(s)"
        self.usage = f"{self.app_name} {self.name}"
        self.env = MMPMEnv()

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

    def exec(self, args, extra):
        if extra:
            logger.error(f"Extra arguments are not accepted. See '{self.app_name} {self.name} --help'")
            return

        logger.debug("Printing user environment to stdout")
        self.env.display()
        MMPMLogFactory.shutdown()



import json
import os
import sys
from datetime import datetime

if sys.version_info.minor > 10:
    from datetime import UTC  # type: ignore

from multiprocessing import cpu_count
from pathlib import Path, PosixPath
from re import sub
from textwrap import fill
from typing import Any, Callable, Dict, List, Tuple

import requests

from mmpm.constants import color
from mmpm.env import MMPMEnv
from mmpm.log.factory import MMPMLogFactory
from mmpm.utils import repo_up_to_date, run_cmd, safe_get_request

NA: str = "N/A"

logger = MMPMLogFactory.get_logger(__name__)


def __sanitize__(string: str) -> str:
    return sub("[//]", "", string)


# pylint: disable=too-many-instance-attributes
class MagicMirrorPackage:
    """
    A container object used to simplify the represenation of a given
    MagicMirror package's metadata
    """

    __slots__ = (
        "title",
        "author",
        "repository",
        "description",
        "category",
        "directory",
        "is_installed",
        "env",
        "is_upgradable",
        "stars",
        "last_updated",
    )

    # pylint: disable=unused-argument,too-many-positional-arguments
    def __init__(
        self,
        title: str = NA,
        author: str = NA,
        repository: str = NA,
        description: str = NA,
        category: str = NA,
        directory: str = "",
        is_installed: bool = False,
        stars: int = 0,
        last_updated: str = NA,
        **kwargs,
    ) -> None:
        """
        Initializes a MagicMirrorPackage instance with the provided metadata.

        Parameters:
            title (str): The title of the package.
            author (str): The author of the package.
            repository (str): The repository URL of the package.
            description (str): The description of the package.
            category (str): The category of the package.
            directory (str): The directory where the package is installed.
            is_installed (bool): A flag indicating whether the package is installed.

        Additional keyword arguments are ignored, but intentionally provided as a means to simplify API interaction.
        """

        self.env = MMPMEnv()
        self.title = __sanitize__(title).strip()
        self.author = __sanitize__(author).strip()
        self.repository = repository.strip()
        self.description = description.strip()
        self.directory = Path(directory.strip())
        self.category = category.strip()
        self.is_installed = is_installed
        self.is_upgradable = False
        self.stars = stars
        self.last_updated = last_updated

    def __str__(self) -> str:
        return str(self.serialize())

    def __repr__(self) -> str:
        return str(self.serialize())

    def __hash__(self) -> int:
        return hash((self.repository.lower(), self.directory.name.lower()))

    def __eq__(self, other) -> bool:
        if other is None:
            return bool(hash(self) == __NULL__)
        else:
            return hash(self) == hash(other)

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    # pylint: disable=too-many-positional-arguments
    def display(
        self,
        detailed: bool = False,
        remote: bool = False,
        title_only: bool = False,
        exclude_installed: bool = False,
        hide_installed_indicator: bool = False,
    ) -> None:
        """
        Displays the package information, optionally with additional details or in a simplified format.

        Parameters:
            detailed (bool): Whether to display detailed information about the package.
            remote (bool): Whether to retrieve additional information from the remote repository.
            title_only (bool): Whether to display only the title of the package.
            exclude_installed (bool): Whether to exclude the package if it is installed.
            hide_installed_indicator (bool): Whether to hide the indicator that shows if the package is installed.

        Returns:
            None
        """
        if exclude_installed and self.is_installed:
            return

        if title_only:
            print(f"{self.title} [installed]" if self.is_installed and not hide_installed_indicator else self.title)
            return

        max_width = 100
        default_fill = lambda message: fill(message, width=max_width, initial_indent="\t", subsequent_indent="\t") + "\n"

        text = color.n_green(self.title) + (" [installed]" if self.is_installed else "") + "\n"

        if detailed:
            modules_dir: PosixPath = self.env.MMPM_MAGICMIRROR_ROOT.get() / "modules"

            text += default_fill(f"Directory: {modules_dir / self.directory}")
            text += default_fill(f"Category: {self.category}")
            text += default_fill(f"Author: {self.author}")
            text += default_fill(f"Repository: {self.repository}")
            text += default_fill(f"Installed: {'true' if self.is_installed else 'false'}")
            text += default_fill(f"Stars: {self.stars}")
            text += default_fill(f"Last Updated: {self.last_updated}")

            if remote:
                for key, value in RemotePackage(self).serialize().items():
                    text += default_fill(f"{key.replace('_', ' ').capitalize()}: {value}")

            text += fill(f"Description: {self.description}", width=max_width, initial_indent="\t", subsequent_indent="\t\t     ") + "\n"
        else:
            text += fill(self.description, width=100, initial_indent="\t", subsequent_indent="\t") + "\n"

        print(f"{text}\n")

    def serialize(self, full: bool = False) -> Dict[str, Any]:
        """
        Serializes the package data into a dictionary.

        Parameters:
            full (bool): If True, returns the full data, otherwise a subset.

        Returns:
            Dict[str, Any]: The serialized package data.
        """

        serialized = {
            "title": self.title,
            "author": self.author,
            "category": self.category,
            "repository": self.repository,
            "description": self.description,
            "directory": self.directory.name,
            "stars": self.stars,
            "last_updated": self.last_updated,
        }

        if full:
            serialized["is_installed"] = self.is_installed  # type: ignore
            serialized["is_upgradable"] = self.is_upgradable  # type: ignore

        return serialized

    def install(self) -> bool:
        """
        Installs the package by cloning the repository and installing dependencies.

        Parameters:
            None

        Returns:
            bool: True if the installation is successful, False otherwise.
        """

        return InstallationHandler(self).install()

    def remove(self) -> bool:
        """
        Removes the package from the installation directory.

        Parameters:
            None

        Returns:
            bool: True if the package is successfully removed, False otherwise.
        """

        modules_dir: PosixPath = self.env.MMPM_MAGICMIRROR_ROOT.get() / "modules"
        error_code, stdout, stderr = run_cmd(["rm", "-rf", str(modules_dir / self.directory)], message="Removing package")
        return not error_code and not stderr and not stdout

    def clone(self) -> Tuple[int, str, str]:
        """
        Clones the package repository into the MagicMirror modules directory.

        Parameters:
            None

        Returns:
            Tuple[int, str, str]: The result of the clone operation including any error codes and messages.
        """

        modules_dir: PosixPath = self.env.MMPM_MAGICMIRROR_ROOT.get() / "modules"

        return run_cmd(
            ["git", "clone", self.repository, str(modules_dir / self.directory)],
            message="Downloading",
        )

    def update(self) -> None:
        """
        Checks for updates to the package by querying the remote repository.

        Parameters:
            None

        Returns:
            None
        """
        modules_dir: PosixPath = self.env.MMPM_MAGICMIRROR_ROOT.get() / "modules"

        if not modules_dir.exists():
            logger.fatal(f"{self.env.MMPM_MAGICMIRROR_ROOT.name}='{str(modules_dir)}' does not exist.")
            self.is_upgradable = False
            return

        os.chdir(modules_dir / self.directory)

        try:
            self.is_upgradable = repo_up_to_date(modules_dir / self.directory)
        except KeyboardInterrupt:
            logger.info("User killed process with CTRL-C")
            sys.exit(127)

    def upgrade(self, force: bool = False) -> bool:
        """
        Upgrades the package by pulling the latest changes from the remote repository.

        Parameters:
            force (bool): If True, forces the upgrade even if the repository is up to date.

        Returns:
            bool: True if the upgrade is successful, False otherwise.
        """
        modules_dir: PosixPath = self.env.MMPM_MAGICMIRROR_ROOT.get() / "modules"

        os.chdir(modules_dir / self.directory)

        error_code, stdout, stderr = run_cmd(["git", "pull"], message="Retrieving changes")

        if error_code or stderr:
            logger.error(f"Failed to upgrade {self.title}: {stderr}")
            return False

        elif "up to date" not in stdout or force and InstallationHandler(self).install():
            print(f"Upgraded {color.n_green(self.title)}")
            logger.debug(f"Upgraded {color.n_green(self.title)}")

        return True

    @classmethod
    def from_json(cls, data: Dict[str, Any]):
        """Create a MagicMirrorPackage from a JSON dict coming from the modules.magicmirror.builders index.

        Expected JSON structure from https://modules.magicmirror.builders/data/modules.json:
        {
          "name": "MMM-ModuleName",
          "category": "Category Name",
          "url": "https://github.com/user/repo",
          "id": "user/repo",
          "maintainer": "username",
          "description": "Module description",
          "stars": <count>,
          "lastCommit": "YYYY-MM-DDTHH:mm:ss+00:00",
          ...
        }

        Parameters:
            data (Dict[str, Any]): The JSON dictionary for a single module

        Returns:
            MagicMirrorPackage: A new instance with data from the JSON
        """
        # Extract fields from the JSON (using the actual field names from modules.json)
        title = data.get("name") or NA
        author = data.get("maintainer") or NA
        repository = data.get("url") or NA
        description = data.get("description") or NA
        category = data.get("category") or NA
        stars = data.get("stars") or 0
        last_commit = data.get("lastCommit") or NA

        if last_commit != NA:
            last_updated = last_commit.split("T")[0]
            # time_format = "%Y-%m-%dT%H:%M:%S%z"
            # date = datetime.strptime(last_commit, time_format)
            # last_updated = f"{date.year}-{date.month}-{date.day}"
        else:
            last_updated = last_commit

        # Derive directory name from the repository URL or id
        directory = ""

        if isinstance(repository, str) and repository and repository != NA:
            directory = repository.split("/")[-1].replace(".git", "")
        elif isinstance(data.get("id"), str) and data.get("id"):
            directory = str(data.get("id")).split("/")[-1]

        return MagicMirrorPackage(
            title=title,
            author=author,
            repository=repository,
            description=description,
            category=category,
            directory=directory,
            last_updated=last_updated,
            stars=stars,
        )


__NULL__: int = hash(MagicMirrorPackage())


class InstallationHandler:
    """
    Delegate class that handles the installation process of
    MagicMirrorPackage's by cloning their repo and identifying dependencies
    that need to be installed.
    """

    __slots__ = {"package"}

    def __init__(self, package: MagicMirrorPackage):
        self.package = package

    def exec(self, funk: Callable) -> bool:
        logger.debug(f"Calling exec wrapper to install dependencies for '{self.package.title}'")
        error_code, _, stderr = funk()

        if error_code:
            logger.error(stderr)
            return False

        return True

    # pylint: disable=too-many-return-statements
    def install(self) -> bool:
        """
        Utility method that detects package.json, Gemfiles, Makefiles, and
        CMakeLists.txt files, and handles the build process for each of the
        previously mentioned files. If the install is successful, an empty string
        is returned. The installation process relies on the location of the current
        directory the os library detects.

        Parameters:
            directory (str): the root directory of the package

        Returns:
            stderr (str): success if the string is empty, fail if not
        """
        root = self.package.env.MMPM_MAGICMIRROR_ROOT
        modules_dir = root.get() / "modules"
        self.package.directory = modules_dir / self.package.directory

        if not modules_dir.exists():
            logger.fatal(f"{root.name}='{modules_dir}' does not exist. Is {root.name} set properly?")
            return False

        os.chdir(modules_dir)

        if not (self.package.directory / ".git").exists():
            logger.debug(f"{self.package.directory / '.git'} not found. Cloning repo.")
            error_code, _, stderr = self.package.clone()

            if error_code:
                logger.error(f"Failed to clone {self.package.title}: {stderr}")
                return False

        os.chdir(self.package.directory)

        if self.exists("package.json"):
            return self.exec(self.npm_install)
        elif self.exists("Gemfile"):
            return self.exec(self.bundle_install)
        elif self.exists("Makefile"):
            return self.exec(self.make)
        elif self.exists("CMakeLists.txt"):
            return self.exec(self.cmake)
        elif self.exists("requirements.txt"):
            return self.exec(self.pip_install)
        elif self.exists("pom.xml"):
            return self.exec(self.maven_install)
        elif self.exists("go.mod"):
            return self.exec(self.go_build)

        logger.debug(f"Unable to find any dependency file associated with {self.package.title}")
        return True

    def cmake(self) -> Tuple[int, str, str]:
        """
        Wrapper method around calling cmake to build a module's dependencies.

        Parameters:
            None

        Returns:
            Tuple[int, str, str]: A tuple containing the exit code, stdout, and stderr from the 'pip install' command.
        """
        logger.debug(f"Running 'cmake ..' in {self.package.directory}")

        build_dir = Path(self.package.directory / "build")
        build_dir.mkdir(exist_ok=True)

        os.system(f"rm -rf {build_dir}/*")
        os.chdir(build_dir)

        return run_cmd(["cmake", ".."], message="Building with CMake")

    def make(self) -> Tuple[int, str, str]:
        """
        Wrapper method around calling make to build a module's dependencies.

        Parameters:
            None

        Returns:
            Tuple[int, str, str]: A tuple containing the exit code, stdout, and stderr from the 'pip install' command.
        """
        logger.debug(f"Found Makefile. Running `make -j {cpu_count()} in {self.package.directory}`")
        return run_cmd(["make", "-j", f"{cpu_count()}"], message="Building with 'make'")

    def npm_install(self) -> Tuple[int, str, str]:
        """
        Wrapper method around calling 'npm install' to install/build a module's dependencies.

        Parameters:
            None

        Returns:
            Tuple[int, str, str]: A tuple containing the exit code, stdout, and stderr from the 'pip install' command.
        """
        logger.debug(f"Found package.json. Running `npm install` in {self.package.directory}")
        return run_cmd(["npm", "install"], message="Installing Node dependencies")

    def bundle_install(self) -> Tuple[int, str, str]:
        """
        Wrapper method around calling 'bundle install' to install/build a module's dependencies.

        Parameters:
            None

        Returns:
            Tuple[int, str, str]: A tuple containing the exit code, stdout, and stderr from the 'pip install' command.
        """
        logger.debug(f"Found Gemfile. Running `bundle install` in {self.package.directory}")
        return run_cmd(["bundle", "install"], message="Installing Ruby dependencies")

    def pip_install(self) -> Tuple[int, str, str]:
        """
        Wrapper method around calling 'pip install' to install/build a module's dependencies.

        Parameters:
            None

        Returns:
            Tuple[int, str, str]: A tuple containing the exit code, stdout, and stderr from the 'pip install' command.
        """
        logger.debug(f"Running 'pip install' in {self.package.directory}")
        return run_cmd(
            ["pip", "install", "-r", "requirements.txt"],
            message="Installing Python dependencies",
        )

    def maven_install(self) -> Tuple[int, str, str]:
        """
        Wrapper method around calling 'maven install' to install/build a module's dependencies.

        Parameters:
            None

        Returns:
            Tuple[int, str, str]: A tuple containing the exit code, stdout, and stderr from the 'pip install' command.
        """
        logger.debug(f"Running 'mvn install' in {self.package.directory}")
        return run_cmd(["mvn", "install"], message="Building with Maven")

    def go_build(self) -> Tuple[int, str, str]:
        """
        Wrapper method around calling 'go install' to install/build a module's dependencies.

        Parameters:
            None

        Returns:
            Tuple[int, str, str]: A tuple containing the exit code, stdout, and stderr from the 'pip install' command.
        """
        logger.debug(f"Running 'go build' in {self.package.directory}")
        return run_cmd(["go", "build"], message="Building Go project")

    def exists(self, file_name: str) -> bool:
        """
        Verifies the dependency file exists in the module's directory.

        Parameters:
            file_name (str): the name of the dependency to look for

        Returns:
            True if the file exists, False otherwise
        """
        return Path(self.package.directory / file_name).exists()


class RemotePackage:
    """
    Class that collects details about a MagicMirrorPackage from its repository.
    """

    def __init__(self, package: MagicMirrorPackage):
        self.package = package

    @classmethod
    def health(cls):
        """
        Checks the health of GitHub, GitLab, and Bitbucket APIs and their rate limits.

        Parameters: None

        Returns:
            dict: A dictionary containing the health status of GitHub, GitLab, and Bitbucket APIs.
        """
        health: dict = {
            "github": {"error": "", "warning": ""},
            "gitlab": {"error": "", "warning": ""},
            "bitbucket": {"error": "", "warning": ""},
        }

        github_api_response: requests.Response = safe_get_request("https://api.github.com/rate_limit")

        if not github_api_response.status_code or github_api_response.status_code != 200:
            health["github"]["error"] = "Unable to contact GitHub API"

        github_api: dict = json.loads(github_api_response.text)
        reset: int = github_api["rate"]["reset"]
        remaining: int = github_api["rate"]["remaining"]

        if sys.version_info.minor > 10:
            reset_time = datetime.fromtimestamp(reset, UTC).strftime("%Y-%m-%d %H:%M:%S")
        else:
            reset_time = datetime.utcfromtimestamp(reset).strftime("%Y-%m-%d %H:%M:%S")

        if not remaining:
            health["github"]["error"] = (
                f"Unable to use `--verbose` option. No GitHub API requests remaining. Request count will reset at {reset_time}"
            )
        elif remaining < 10:
            health["github"]["warning"] = f"{remaining} GitHub API requests remaining. Request count will reset at {reset_time}"

        try:
            # GitLab doesn't have rate limits that will cause any issues with checking for repos
            gitlab_api = requests.head("https://gitlab.com", allow_redirects=True, timeout=10)

            if gitlab_api.status_code != 200:
                health["gitlab"]["error"] = "GitLab server returned invalid response"
        except requests.exceptions.RequestException:
            health["gitlab"]["error"] = "Unable to communicate with GitLab server"

        try:
            # Bitbucket rate limits are similar to GitLab
            bitbucket_api = requests.head("https://bitbucket.org", allow_redirects=True, timeout=10)

            if bitbucket_api.status_code != 200:
                health["bitbucket"]["error"] = "Bitbucket server returned invalid response"
        except requests.exceptions.RequestException:
            health["gitlab"]["error"] = "Unable to communicate with Bitbucket server"

        return health

    def serialize(self):
        """
        Retrieves and formats details about the MagicMirror package from its remote repository.

        Parameters: None

        Returns:
            dict: A dictionary containing details such as forks, issue counts, and creation and last updated dates of the repository.
        """
        spliced: List[str] = self.package.repository.split("/")
        user: str = spliced[-2]
        project: str = spliced[-1].replace(".git", "")  # in case the user added .git to the end of the url
        details = {}

        if "github" in self.package.repository:
            url = f"https://api.github.com/repos/{user}/{project}"
            logger.debug(f"Constructed {url} to request more details for {self.package.title}")
            data = safe_get_request(url)

            if not data:
                logger.error(f"Unable to retrieve {self.package.title} details, data was empty")

            details = self.__format_github_api_details__(json.loads(data.text)) if data else {}

        elif "gitlab" in self.package.repository:
            url = f"https://gitlab.com/api/v4/projects/{user}%2F{project}"
            logger.debug(f"Constructed {url} to request more details for {self.package.title}")
            data = safe_get_request(url)

            if not data:
                logger.error(f"Unable to retrieve {self.package.title} details, data was empty")

            details = self.__format_gitlab_api_details__(json.loads(data.text), url) if data else {}

        elif "bitbucket" in self.package.repository:
            url = f"https://api.bitbucket.org/2.0/repositories/{user}/{project}"
            logger.debug(f"Constructed {url} to request more details for {self.package.title}")
            data = safe_get_request(url)

            if not data:
                logger.error(f"Unable to retrieve {self.package.title} details, data was empty")

            details = self.__format_bitbucket_api_details__(json.loads(data.text), url) if data else {}

        return details

    def __format_bitbucket_api_details__(self, data: dict, url: str) -> dict:
        """
        Helper method to format remote repository data from Bitbucket.

        Parameters:
            data (dict): JSON data from the API request.
            url (str): The constructed URL of the API used to retrieve additional info.

        Returns:
            dict: A dictionary with forks, issue counts, and creation and last updated dates.
        """
        forks = safe_get_request(f"{url}/forks")
        issues = safe_get_request(f"{url}/issues")

        return (
            {
                "issues": int(json.loads(issues.text)["pagelen"]) if issues else "N/A",
                "created": data["created_on"].split("T")[0] if data else "N/A",
                "forks": int(json.loads(forks.text)["pagelen"]) if forks else "N/A",
            }
            if data
            else {}
        )

    def __format_gitlab_api_details__(self, data: dict, url: str) -> dict:
        """
        Helper method to format remote repository data from GitLab.

        Parameters:
            data (dict): JSON data from the API request.
            url (str): The constructed URL of the API used to retrieve additional info.

        Returns:
            dict: A dictionary with forks, issue counts, and creation and last updated dates.
        """
        issues = safe_get_request(f"{url}/issues")

        return (
            {
                "issues": len(json.loads(issues.text)) if issues else "N/A",
                "created": data["created_at"].split("T")[0] if data else "N/A",
                "forks": data["forks_count"] if data else "N/A",
            }
            if data
            else {}
        )

    def __format_github_api_details__(self, data: dict) -> dict:
        """
        Helper method to format remote repository data from GitHub.

        Parameters:
            data (dict): JSON data from the API request.

        Returns:
            dict: A dictionary with forks, issue counts, and creation and last updated dates.
        """
        return (
            {
                "issues": data["open_issues"] if data else "N/A",
                "created": data["created_at"].split("T")[0] if data else "N/A",
                "forks": data["forks_count"] if data else "N/A",
            }
            if data
            else {}
        )



import datetime
import json
import logging
import logging.handlers
import os
import zipfile
from pathlib import Path
from threading import Lock

import socketio

from mmpm.__version__ import version
from mmpm.constants import paths
from mmpm.env import MMPMEnv


class JsonFormatter(logging.Formatter):
    """
    A custom formatter for logging, which outputs log records in a JSON format.
    """

    def format(self, record):
        """
        Formats the log record into JSON.

        Parameters:
            record (logging.LogRecord): The log record to be formatted.

        Returns:
            str: A JSON string representation of the log record.
        """

        try:
            message = record.getMessage()
        except TypeError:
            # Handling the case where formatting fails
            message = {"raw_message": record.msg, "args": record.args}

        log_data = {
            "timestamp": self.formatTime(record),
            "level": record.levelname,
            "version": version,
            "message": message,
            "logger_name": record.name,
            "module": record.module,
            "function": record.funcName,
            "line": record.lineno,
        }

        return json.dumps(log_data, ensure_ascii=False)


class SocketIOHandler(logging.Handler):
    """A logging handler that emits records with SocketIO."""

    def __init__(self, host, port):
        """
        Initializes the SocketIOHandler with a specified host and port for the SocketIO server.

        Parameters:
            host (str): The host name of the SocketIO server.
            port (int): The port number of the SocketIO server.
        """

        super().__init__()
        self.formatter = JsonFormatter()
        self.sio = socketio.Client()

        try:
            self.sio.connect(f"http://{host}:{port}", wait=False)
        except socketio.exceptions.ConnectionError:
            pass

    def emit(self, record):
        """
        Emits the log record to the connected SocketIO server.

        Parameters:
            record (logging.LogRecord): The log record to be emitted.
        """

        if self.sio.connected:
            try:
                self.sio.emit("logs", self.formatter.format(record))
            except Exception:
                pass

    def close(self):
        """
        Closes the connection to the SocketIO server and performs any necessary cleanup.

        Parameters:
            None
        """

        if self.sio.connected:
            self.sio.disconnect()
            super().close()


class StdoutFormatter(logging.Formatter):
    """
    A custom formatter for logging, which outputs log records to stdout with a simplified format.
    """

    def format(self, record):
        """
        Formats the log record for stdout.

        Parameters:
            record (logging.LogRecord): The log record to be formatted.

        Returns:
            str: A string representation of the log record for stdout.
        """

        label = "+" if record.levelname == "INFO" else record.levelname

        return f"[{label}] {record.getMessage()}"


class MMPMLogFactory:
    """
    A custom logging class for MMPM, providing functionalities for logging to files, stdout, and SocketIO.
    Logs can be found in ~/.config/mmpm/log.
    """

    __logger: logging.Logger = None
    __socketio_handler: SocketIOHandler = None
    __lock: Lock = Lock()

    @staticmethod
    def __setup__(name: str) -> None:
        MMPMLogFactory.__logger = logging.getLogger(name)

        file_handler = logging.handlers.RotatingFileHandler(
            paths.MMPM_CLI_LOG_FILE,
            mode="a",
            maxBytes=1024 * 1024,
            backupCount=2,
            encoding="utf-8",
            delay=False,
        )

        level = MMPMEnv().MMPM_LOG_LEVEL.get()
        MMPMLogFactory.__logger.setLevel(logging.DEBUG)  # set the main logging handler set to the lowest level possible

        file_handler.setFormatter(JsonFormatter())
        file_handler.setLevel(logging.DEBUG)  # always have the log files be DEBUG
        MMPMLogFactory.__logger.addHandler(file_handler)

        stdout_handler = logging.StreamHandler()
        stdout_handler.setFormatter(StdoutFormatter())
        stdout_handler.setLevel(level)

        MMPMLogFactory.__logger.addHandler(stdout_handler)

        socketio_handler = SocketIOHandler("localhost", 6789)

        if socketio_handler.sio.connected:
            MMPMLogFactory.__socketio_handler = socketio_handler
            MMPMLogFactory.__socketio_handler.setLevel(logging.DEBUG)
            MMPMLogFactory.__logger.addHandler(MMPMLogFactory.__socketio_handler)
        else:
            MMPMLogFactory.__logger.debug("Failed to connect to SocketIO server")

    @staticmethod
    def shutdown() -> None:
        """
        Shuts down the logger, closing any SocketIO connections.

        Parameters:
            None
        """

        if MMPMLogFactory.__socketio_handler is not None:
            MMPMLogFactory.__logger.debug("Disconnecting from SocketIO server")
            MMPMLogFactory.__socketio_handler.close()

    @staticmethod
    def get_logger(name: str) -> logging.Logger:
        """
        Retrieves a logger instance with the given name, initializing it if necessary.

        Parameters:
            name (str): The name of the logger.

        Returns:
            logging.Logger: The logger instance associated with the given name.
        """

        if MMPMLogFactory.__logger is None:
            with MMPMLogFactory.__lock:
                if MMPMLogFactory.__logger is None:
                    MMPMLogFactory.__setup__(name)

        return MMPMLogFactory.__logger

    @classmethod
    def display(cls, tail: bool = False) -> None:
        """
        Displays contents of log files to stdout. If the tail option is supplied, log contents will be displayed in real-time.

        Parameters:
            tail (bool): If True, displays the log contents in real time.

        Returns:
            None
        """
        if paths.MMPM_CLI_LOG_FILE.exists():
            os.system(f"{'tail -F' if tail else 'cat'} {paths.MMPM_CLI_LOG_FILE}")
        else:
            MMPMLogFactory.__logger.error("MMPM log file not found")

    @classmethod
    def archive(cls) -> None:
        """
        Compresses all log files in ~/.config/mmpm/log, excluding NGINX logs.

        Parameters:
            None

        Returns:
            None
        """
        today = datetime.datetime.now()
        file_name = Path(f"mmpm-logs-{today.year}-{today.month}-{today.day}.zip")

        try:
            with zipfile.ZipFile(file_name, mode="w") as archive:
                os.chdir(paths.MMPM_LOG_DIR)

                for file in os.listdir(os.getcwd()):
                    archive.write(file)

                if paths.PM2_LOG_DIR.exists():
                    os.chdir(paths.PM2_LOG_DIR)

                    for file in os.listdir(os.getcwd()):
                        archive.write(file)

        except (IOError, zipfile.BadZipFile, Exception) as error:
            MMPMLogFactory.__logger.error(f"Failed to create archive of log files: {error}")
            return

        MMPMLogFactory.__logger.info(f"Compressed MMPM log files to {os.getcwd()}/{file_name}")



from flask import Blueprint, Response

from mmpm.api.constants import http
from mmpm.api.endpoints.endpoint import Endpoint
from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.database import MagicMirrorDatabase
from mmpm.magicmirror.magicmirror import MagicMirror
from mmpm.utils import update_available

logger = MMPMLogFactory.get_logger(__name__)


class Db(Endpoint):
    """
    Endpoint to interact with the MagicMirrorDatabase object. Handles updating, upgrading, and collecting information about the database.
    """

    def __init__(self):
        self.name = "db"
        self.blueprint = Blueprint(self.name, __name__, url_prefix=f"/api/{self.name}")
        self.db = MagicMirrorDatabase()
        self.magicmirror = MagicMirror()

        @self.blueprint.route("/update", methods=[http.GET])
        def update() -> Response:
            """
            A Flask route method for updating the MagicMirror database.

            Parameters:
                None

            Returns:
                Response: A Flask Response object indicating the success or failure of the database update.
            """
            if not self.db.load(update=True):
                return self.failure("Failed to update database. See logs for details.")

            can_upgrade_mmpm = update_available()
            can_upgrade_magicmirror = self.magicmirror.update()
            self.db.update(can_upgrade_mmpm=can_upgrade_mmpm, can_upgrade_magicmirror=can_upgrade_magicmirror)

            return self.success(self.db.upgradable())

        @self.blueprint.route("/upgradable", methods=[http.GET])
        def upgradable() -> Response:
            """
            A Flask route method for retrieving a list of upgradable packages from the MagicMirror database.

            Parameters:
                None

            Returns:
                Response: A Flask Response object containing a list of upgradable packages or an error message.
            """

            upgradable = self.db.upgradable()

            if upgradable:
                return self.success(upgradable)

            return self.failure("Failed to get list of upgradable packages. See logs for details")

        @self.blueprint.route("/info", methods=[http.GET])
        def info() -> Response:
            """
            A Flask route method for retrieving information about the MagicMirror database.

            Parameters:
                None

            Returns:
                Response: A Flask Response object containing database information or an error message.
            """

            info = self.db.info()

            if info:
                return self.success(info)

            return self.failure("Failed to retrieve database info. See logs for details")



import json
from os.path import getmtime
from pathlib import Path

from pygments import highlight
from pygments.formatters.terminal import TerminalFormatter
from pygments.lexers.data import JsonLexer

from mmpm.constants import color, paths
from mmpm.singleton import Singleton

MMPM_DEFAULT_ENV: dict = {
    "MMPM_MAGICMIRROR_ROOT": Path(paths.HOME_DIR / "MagicMirror"),
    "MMPM_MAGICMIRROR_URI": "http://localhost:8080",
    "MMPM_MAGICMIRROR_PM2_PROCESS_NAME": "",
    "MMPM_MAGICMIRROR_DOCKER_COMPOSE_FILE": "",
    "MMPM_IS_DOCKER_IMAGE": False,
    "MMPM_LOG_LEVEL": "INFO",
}


class EnvVar:
    """
    Represents a re-readable environment variable stored in the MMPM_ENV_FILE.
    When a change in the last-modified time of the MMPM_ENV_FILE, the value is
    re-read. The __slots__ are predefined to improve efficiency.

    Attributes:
        name (str): the name of the environment variable
        default (Any): the default value
        __tipe (object): the class the environment should be initialized as
        __mtime (object): the last modified time of the MMPM_ENV_FILE
        __value (object): the value read from the MMPM_ENV_FILE

    Methods:
        get(): Returns the value of the environment variable
    """

    __slots__ = "name", "default", "__tipe", "__value", "__mtime"

    def __init__(self, name: str = "", default=None, mtime: float = None):
        self.name: str = name
        self.default = default
        self.__tipe = type(default)  # avoid name clashing with 'type'
        self.__mtime: float = mtime
        self.__value = None

    def get(self):
        """
        Reads environment variables from the MMPM_ENV_FILE. In order to ensure
        hot-reloading is usable in the UI, the environment variables need to be
        re-read from the file each time. Otherwise, cached data will be sent back
        to the user.

        Parameters:
            None

        Returns:
            value:  the value of the environment variable key
        """

        mtime: float = getmtime(paths.MMPM_ENV_FILE)

        if mtime != self.__mtime or self.__value is None:  # cache the value until the file modification time changes
            with open(paths.MMPM_ENV_FILE, "r", encoding="utf-8") as env:
                env_vars = {}

                try:
                    env_vars = json.load(env)
                except json.JSONDecodeError:
                    print(
                        color.b_yellow("WARNING:"),
                        "Unable to parse environment variables file.",
                    )

                # make sure we construct the expected type using from parsed data, otherwise instead of
                # something like a Path object we would return a string
                self.__value = self.__tipe(self.default if self.name not in env_vars else env_vars.get(self.name))

            self.__mtime = mtime

        return self.__value


class MMPMEnv(Singleton):
    """
    MMPMEnv, a singleton class, serves as the centralized source for managing and accessing environment variables
    within the MMPM application. It reads and writes environment variables to the MMPM_ENV_FILE, ensuring that all
    components of the application have consistent and up-to-date configurations. The class is designed to dynamically
    reflect changes made to the environment variables in the MMPM_ENV_FILE.

    Attributes:
        MMPM_MAGICMIRROR_ROOT (EnvVar): Environment variable for the root directory of MagicMirror.
        MMPM_MAGICMIRROR_URI (EnvVar): Environment variable for the URI of the MagicMirror.
        MMPM_MAGICMIRROR_PM2_PROCESS_NAME (EnvVar): Environment variable for the PM2 process name of MagicMirror.
        MMPM_MAGICMIRROR_DOCKER_COMPOSE_FILE (EnvVar): Environment variable for the Docker compose file path.
        MMPM_IS_DOCKER_IMAGE (EnvVar): Environment variable indicating if MMPM is running as a Docker image.
        MMPM_LOG_LEVEL (EnvVar): Environment variable for the logging level.

    Methods:
        __init__(): Initializes the MMPMEnv instance, loading environment variables from MMPM_ENV_FILE.
        get(): Retrieves the current environment variables as a dictionary.
        display(): Prints the current environment variables in a formatted JSON structure for easy viewing.
    """

    __slots__ = tuple({key.lower() for key in MMPM_DEFAULT_ENV})

    def __init__(self):
        super().__init__()
        self.MMPM_MAGICMIRROR_ROOT: EnvVar = None
        self.MMPM_MAGICMIRROR_URI: EnvVar = None
        self.MMPM_MAGICMIRROR_PM2_PROCESS_NAME: EnvVar = None
        self.MMPM_MAGICMIRROR_DOCKER_COMPOSE_FILE: EnvVar = None
        self.MMPM_IS_DOCKER_IMAGE: EnvVar = None
        self.MMPM_LOG_LEVEL: EnvVar = None

        env_vars = {}

        with open(paths.MMPM_ENV_FILE, "r", encoding="utf-8") as env_file:
            try:
                env_vars = json.load(env_file)
            except json.JSONDecodeError:
                pass

            for key, value in MMPM_DEFAULT_ENV.items():
                if key not in env_vars:
                    env_vars[key] = value

        env_vars["MMPM_MAGICMIRROR_ROOT"] = str(env_vars["MMPM_MAGICMIRROR_ROOT"])

        with open(paths.MMPM_ENV_FILE, "w", encoding="utf-8") as env:
            json.dump(env_vars, env, indent=2)

        mtime: float = getmtime(paths.MMPM_ENV_FILE)

        for key, value in MMPM_DEFAULT_ENV.items():
            if hasattr(self, key):
                setattr(self, key, EnvVar(name=key, default=value, mtime=mtime))

    def get(self) -> dict:
        current_env: dict = {}

        with open(paths.MMPM_ENV_FILE, "r", encoding="utf-8") as env:
            current_env = json.load(env)

        return current_env

    def display(self) -> None:  # pragma: no cover
        print(highlight(json.dumps(self.get(), indent=2), JsonLexer(), TerminalFormatter()))



from mmpm.utils import get_host_ip

HOST = f"{get_host_ip()}"

MMPM_UI_PORT = 7890
MMPM_API_SERVER_PORT = 7891
MMPM_LOG_SERVER_PORT = 6789
MMPM_REPEATER_SERVER_PORT = 8907

MMPM_WIKI_URL = "https://github.com/Bee-Mar/mmpm/wiki"

MAGICMIRROR_REPO_URL: str = "https://github.com/MagicMirrorOrg/MagicMirror"
MAGICMIRROR_WIKI_URL: str = f"{MAGICMIRROR_REPO_URL}/wiki"
MAGICMIRROR_DOCUMENTATION_URL: str = "https://docs.magicmirror.builders/"
MAGICMIRROR_MODULES_URL: str = "https://modules.magicmirror.builders/data/modules.json"



import sys
from argparse import ArgumentParser

import argcomplete

import mmpm.subcommands
from mmpm.constants import urls
from mmpm.log.factory import MMPMLogFactory
from mmpm.subcommands.loader import Loader

logger = MMPMLogFactory.get_logger(__name__)


# for the console script
def main():
    """
    Initializes all subcommand and their options.

    Parameters:
        None

    Returns:
        None
    """

    if "--help" in sys.argv or "-h" in sys.argv:
        # close up any SocketIO connection in the logger that could linger before it becomes a problem
        MMPMLogFactory.shutdown()

    app_name = "mmpm"

    parser = ArgumentParser(
        prog=app_name,
        usage=f"{app_name} <subcommand> [option(s)]",
        epilog=f"Visit {urls.MMPM_WIKI_URL} for more details",
        description="""
            The MagicMirror Package Manager (MMPM) CLI simplifies the
            installation, removal, and general maintenance of MagicMirror packages.
            """,
    )

    subparser = parser.add_subparsers(
        title=f"{app_name} subcommands",
        description=f"use `{app_name} <subcommand> --help` to see more details",
        dest="subcmd",
        metavar="",
    )

    loader = Loader(
        module_path=mmpm.subcommands.__path__,
        module_name="mmpm.subcommands",
        app_name=app_name,
        prefix="_sub_cmd",
    )

    for subcommand in loader.objects.values():
        subcommand.register(subparser)

    argcomplete.autocomplete(parser)

    args, extra = parser.parse_known_args()
    subcommand = loader.objects.get(args.subcmd)

    if not subcommand:
        logger.debug(f"Unable to match '{args.subcmd}' to a valid subcommand")
        parser.print_help()
    else:
        subcommand.exec(args, extra)

    MMPMLogFactory.shutdown()


if __name__ == "__main__":
    main()



"""Command line options for 'upgrade' subcommand"""

import json
from typing import List

from mmpm import utils
from mmpm.constants import paths
from mmpm.env import MMPMEnv
from mmpm.log.factory import MMPMLogFactory
from mmpm.magicmirror.database import MagicMirrorDatabase
from mmpm.magicmirror.magicmirror import MagicMirror
from mmpm.magicmirror.package import MagicMirrorPackage
from mmpm.subcommands.sub_cmd import SubCmd

logger = MMPMLogFactory.get_logger(__name__)


class Upgrade(SubCmd):
    """
    The 'Upgrade' subcommand retrieves available updates for packages and MagicMirror and attempts to install their dependencies

    Custom Attributes:
        database (MagicMirrorDatabase): An instance of the MagicMirrorDatabase class for managing the database.
        magicmirror (MagicMirror): An instance of the MagicMirror object (similar to a MagicMirrorPackage)
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "upgrade"
        self.help = "Upgrade packages, MMPM, and/or MagicMirror"
        self.usage = f"{self.app_name} {self.name} <package(s)> [--yes]"
        self.database = MagicMirrorDatabase()
        self.magicmirror = MagicMirror()
        self.env = MMPMEnv()

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

        self.parser.add_argument(
            "-y",
            "--yes",
            action="store_true",
            default=False,
            help="assume yes for user response and do not show prompt",
            dest="assume_yes",
        )

        self.parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            default=False,
            help="force an attempted upgrade regardless if the package isn't 'upgradable'",
            dest="force",
        )

    def exec(self, args, extra):
        if not self.database.is_initialized():
            self.database.load()

        upgradable = self.database.upgradable()
        packages_to_upgrade: List[MagicMirrorPackage] = []

        if args.force:
            for package in filter(lambda pkg: pkg.is_installed, self.database.packages):
                package.upgrade(force=True)

            upgradable["packages"] = {}

        elif not any(upgradable.values()):
            logger.info("All packages and applications are up to date.\n ")
            return

        if upgradable["packages"]:
            packages = {MagicMirrorPackage(**package) for package in upgradable["packages"]}
            packages_to_upgrade.extend(filter(lambda pkg: pkg.upgrade(), packages))
            upgradable["packages"] = [package.serialize() for package in (packages - set(packages_to_upgrade))]

        upgradable["MagicMirror"] = upgradable["MagicMirror"] and self.magicmirror.upgrade()

        if upgradable["mmpm"]:
            if self.env.MMPM_IS_DOCKER_IMAGE.get():
                logger.warning("Cannot perform self-upgrade because MMPM is a Docker image. Stop MMPM and run `docker pull karsten13/mmpm:latest`")
            else:
                upgradable["mmpm"] = utils.upgrade()

        with open(paths.MMPM_AVAILABLE_UPGRADES_FILE, mode="w", encoding="utf-8") as upgrade_file:
            json.dump(upgradable, upgrade_file)



"""Command line options for 'completion' subcommand"""

from typing import Dict

from mmpm.log.factory import MMPMLogFactory
from mmpm.subcommands.sub_cmd import SubCmd

logger = MMPMLogFactory.get_logger(__name__)


class Completion(SubCmd):
    """
    The 'Completion' subcommand generates the necessary commands a user should copy/paste into their
    shell configuration file to register autocompletion for MMPM.

    Custom Attributes:
        shells (Dict[str, str]): a dictionary of shell names and associated commands for autocompletion registration
    """

    def __init__(self, app_name):
        self.app_name = app_name
        self.name = "completion"
        self.help = f"Generate commands to enable autocompletion for {self.app_name}"
        self.usage = f"{self.app_name} {self.name} --shell=[type]"

        self.shells: Dict[str, str] = {
            "bash": f'eval "$(register-python-argcomplete {self.app_name})"',
            "zsh": f'autoload -U bashcompinit\nbashcompinit\neval "$(register-python-argcomplete {self.app_name})"',
            "tcsh": f"`register-python-argcomplete --shell tcsh {self.app_name}`",
            "fish": f"register-python-argcomplete --shell fish {self.app_name}",
        }

    def register(self, subparser):
        self.parser = subparser.add_parser(self.name, usage=self.usage, help=self.help)

        self.parser.add_argument(
            "-s",
            "--shell",
            choices=self.shells.keys(),
            help="The shell type to generate completion commands for",
            dest="shell",
        )

    def exec(self, args, extra):
        """
        Adds autocompletion configuration to a user's shell configuration file.
        Detects configuration files for bash, zsh, fish, and tcsh

        Parameters:
            assume_yes (bool): if True, assume yes for user response, and do not display prompt

        Returns:
            None
        """

        if not args.shell or args.shell is None:
            logger.error(f"No shell type provided. See `{self.app_name} {self.name} --help`")
            return

        if args.shell not in self.shells:
            logger.error(f"Invalid shell type. See `{self.app_name} {self.name} --help`")
            return

        logger.debug(f"Outputting commands to enable cli completion for {args.shell}")

        print(f"# enables {self.app_name} tab-completion for subcommands")
        print(self.shells.get(args.shell))



from importlib import import_module
from pkgutil import iter_modules
from typing import Dict

from mmpm.log.factory import MMPMLogFactory

logger = MMPMLogFactory.get_logger(__name__)


class Loader:
    """
    This class handles dynamically loading all subcommands/endpoints from given modules.

    Attributes:
        objects (Dict[str, object]): A dictionary holding the loaded objects with their names as keys.

    Args:
        module_path (str): The file system path to the module to load objects from.
        module_name (str): The name of the module from which objects are to be loaded.
        app_name (str, optional): The name of the app, if applicable. Defaults to an empty string.
        prefix (str, optional): A prefix to filter which submodules to load. Defaults to an empty string.
    """

    def __init__(self, module_path, module_name: str, app_name: str = "", prefix: str = ""):
        """
        Initializes the Loader instance and loads objects from the specified module.

        Args:
            module_path (str): The file system path to the module to load objects from.
            module_name (str): The name of the module from which objects are to be loaded.
            app_name (str, optional): The name of the app, if applicable. Defaults to an empty string.
            prefix (str, optional): A prefix to filter which submodules to load. Defaults to an empty string.
        """

        self.objects: Dict[str, object] = self.__load__(module_path, module_name, app_name, prefix)

    def __load__(self, module_path, module_name: str, app_name: str = "", prefix: str = "") -> Dict[str, object]:
        """
        Loads objects dynamically from the specified module.

        This method scans the given module path for submodules that start with the specified prefix,
        imports them, and creates instances of the classes found within those modules.

        Args:
            module_path (str): The file system path to the module to load objects from.
            module_name (str): The name of the module from which objects are to be loaded.
            app_name (str, optional): The name of the app, if applicable. Used to instantiate the objects if required.
            prefix (str, optional): A prefix to filter which submodules to load. Defaults to an empty string.

        Returns:
            Dict[str, object]: A dictionary of the loaded objects, with object names as keys.

        Raises:
            AttributeError: If a required attribute is missing in the loaded module.
            AssertionError: If an assertion fails in the process of loading.
            Exception: For any other exceptions encountered during loading.
        """

        objects: Dict[str, object] = {}
        snake_to_pascal = lambda name: name.replace("_", " ").title().replace(" ", "")

        for submodule in iter_modules(module_path):
            if submodule.name.startswith(prefix):
                class_name = snake_to_pascal(submodule.name.replace(prefix, ""))

                try:
                    imported_module = import_module(f"{module_name}.{submodule.name}")
                    objekt = getattr(imported_module, class_name)
                    instance = objekt(app_name) if app_name else objekt()
                    objects[instance.name] = instance
                except (AttributeError, AssertionError, Exception) as error:
                    logger.error(f"Failed to load subcommand module: {error}")

        return objects
